"""
Module containing the OptionHolder class
"""

from typing import Optional

from aiida.orm import Dict
from pydantic import BaseModel, Field, ValidationError

# pylint:disable=raise-missing-from


class OptionContainer(BaseModel):
    """
    Base class for a container of options
    """

    def aiida_dict(self) -> Dict:
        """Return an ``aiida.orm.Dict`` presentation"""

        python_dict = self.model_dump()
        return Dict(dict=python_dict)

    @classmethod
    def aiida_validate(cls, input_dict: dict | Dict, namespace: None | str = None) -> None:  # pylint:disable=unused-argument
        """
        Validate a dictionary/Dict node, this can be used as the validator for
        the Port accepting the inputs

        This is used as validator for the `spec.input` call.
        """
        if isinstance(input_dict, Dict):
            input_dict = input_dict.get_dict()
        try:
            cls(**input_dict)
        except ValidationError as error:
            return str(error)
        return None

    @classmethod
    def aiida_serialize(cls, python_dict: dict) -> Dict:
        """
        serialize a dictionary into Dict

        This method can be passed as a `serializer` key word parameter of for the `spec.input` call.
        """
        obj = cls(**python_dict)
        return obj.aiida_dict()

    @classmethod
    def aiida_description(cls) -> str:
        """
        Return a string for the options of a OptionContains in a human-readable format.
        """

        obj = cls()
        template = '{:>{width_name}s}:  {:10s} \n{default:>{width_name2}}: {}'
        entries = []
        for name, field in obj.model_fields.items():
            # Each entry is name, type, doc, default value
            entries.append([name, str(field.annotation.__name__), field.description, field.default])
        max_width_name = max(len(entry[0]) for entry in entries) + 2

        lines = []
        for entry in entries:
            lines.append(
                template.format(
                    *entry,
                    width_name=max_width_name,
                    width_name2=max_width_name + 10,
                    default='Default',
                )
            )
        return '\n'.join(lines)


class CalcSettingsConfig(OptionContainer):
    """Schema for the .settings port used by both VaspCalculation and VaspWorkChain"""

    parser_setting: Optional[dict] = Field(description='Settings for the parser')
    ADDITIONAL_RETRIEVE_LIST: Optional[list] = Field(description='Additional list of files to be retrieved')
    ADDITIONAL_RETRIEVE_TEMPORARY_LIST: Optional[list] = Field(
        description=('Additional list of files to be retrieved, but not store in the storage')
    )
    PROVENANCE_EXCLUDE_LIST: Optional[list] = Field(
        description=('Additional list of files to be retrieved, but not store in the storage')
    )
    ALWAYS_STORE: Optional[list] = Field(
        description=('Additional list of files to be retrieved, but not store in the storage')
    )
    skip_param_validation: Optional[bool] = Field(
        description='Skip the validation of the input parameters', default=False
    )
    unsupported_parameters: Optional[dict] = Field(description='None-standard VASP parameters that are valid')


class RelaxOptions(OptionContainer):
    """Options for VaspRelaxWorkChain"""

    algo: str = Field(description='The algorithm to use for relaxation', examples=['cg', 'rd'], default='cg')
    energy_cutoff: Optional[float] = Field(
        description='The cut off energy difference when the relaxation is stopped (e.g. EDIFF)',
        default=None,
    )
    force_cutoff: float = Field(
        description='The maximum force when the relaxation is stopped (e.g. EDIFFG)',
        default=0.03,
    )
    steps: int = Field(description='Number of relaxation steps to perform (eg. NSW)', default=60)
    positions: bool = Field(description='If True, perform relaxation of the atomic positions', default=True)
    shape: bool = Field(description='If True, perform relaxation of the cell shape', default=True)
    volume: bool = Field(description='If True, perform relaxation of the cell volume', default=True)
    convergence_on: bool = Field(description='If True, perform convergence checks within the workchain', default=True)
    convergence_absolute: bool = Field(
        description='If True, use absolute values where possible when performing convergence checks',
        default=False,
    )
    convergence_max_iterations: int = Field(description='Maximum iterations for convergence checking', default=5)
    convergence_positions: float = Field(
        description=(
            'The cutoff value for the convergence check on positions in Angstram. A negative value by pass the check.'
        ),
        default=0.1,
    )
    convergence_volume: float = Field(
        description='The cutoff value for the convergence check on volume between the two structures.'
        ' A negative value by pass the check.',
        default=0.01,
    )
    convergence_shape_lengths: float = Field(
        description='The cutoff value for the convergence check on the lengths of the unit cell'
        ' vectors, between input and the outputs structure. A negative value by pass'
        ' the check.',
        default=0.1,
    )
    convergence_shape_angles: float = Field(
        description='The cutoff value for the convergence check on the angles of the unit cell vectors, '
        'between input and the outputs structure. A'
        ' negative value by pass the check.',
        default=0.1,
    )
    convergence_mode: str = Field(
        description="Mode of the convergence check for positions. 'inout' for checking input/output structure, "
        "or 'last' to check only the change of"
        ' the last step.',
        examples=['inout', 'last'],
        default='last',
    )
    reuse: bool = Field(
        description='Whether reuse the previous calculation by copying over the remote folder',
        default=False,
    )
    clean_reuse: bool = Field(
        description='Whether to perform a final cleaning of the reused calculations', default=True
    )
    keep_sp_workdir: bool = Field(
        description='Whether to keep the workdir of the final singlepoint calculation', default=False
    )
    perform: bool = Field(description="Do not perform any relaxation if set to 'False'", default=True)
    hybrid_calc_bootstrap: bool = Field(
        description='Whether to bootstrap hybrid calculation by performing standard DFT first', default=False
    )
    hybrid_calc_bootstrap_wallclock: int = Field(
        description='Wall time limit in second for the bootstrap calculation', default=3600
    )
    keep_magnetization: bool = Field(
        description='Whether to keep magnetization from the previous calculation if possible', default=False
    )
    double_relax_mode: bool = Field(
        description='Experimental: Run in double relax mode - launch of the sub workflow is only performed up to two '
        'times without checking convergence in the end. This is useful for cases where the convergence is difficult '
        'due to change of basis set with variable cell and high-throughput studies.',
        default=False,
    )

    # TODO: implement pyandtic checks

    # @classmethod
    # def validate_dict(cls, input_dict, port=None):
    #     """Check mutually exclusive fields"""
    #     super().validate_dict(input_dict, port)
    #     if isinstance(input_dict, orm.Dict):
    #         input_dict = input_dict.get_dict()
    #     force_cut = input_dict.get('force_cutoff')
    #     energy_cut = input_dict.get('energy_cutoff')
    #     if force_cut is None and energy_cut is None:
    #         raise InputValidationError("Either 'force_cutoff' or 'energy_cutoff' should be supplied")
    #     if (force_cut is not None) and (energy_cut is not None):
    #         raise InputValidationError("Cannot set both 'force_cutoff' and 'energy_cutoff'")


class ConvOptions(OptionContainer):
    """Template for the Dict node controlling the workchain behaviour"""

    cutoff_start: float = Field(description='The starting cut-off energy', default=300.0)
    cutoff_stop: float = Field(description='The Final cut-off energy', default=700.0)
    cutoff_step: float = Field(description='Step size of the cut-off energy', default=50.0)
    kspacing_start: float = Field(description='The starting kspacing', default=0.07)
    kspacing_stop: float = Field(description='The final kspacing', default=0.02)
    kspacing_step: float = Field(description='Step size of the cut-off energy', default=0.01)
    cutoff_kconv: float = Field(description='The cut-off energy used for kpoints convergence tests', default=450.0)
    kspacing_cutconv: float = Field(
        description='The kpoints spacing used for cut-off energy convergence tests', default=0.07
    )


class BandOptions(OptionContainer):
    """Options for VaspRelaxWorkChain"""

    symprec: float = Field(description='Precision of the symmetry determination', default=0.01)
    band_mode: str = Field(
        description=(
            'Mode for generating the band path. Choose from: bradcrack, pymatgen,seekpath-aiida and latimer-munro.'
        ),
        examples=['bradcrack', 'pymatgen', 'seekpath', 'seekpath-aiida', 'latimer-munro'],
        default='seekpath-aiida',
    )
    # TODO: enable explicit seekpath passing
    band_kpoints_distance: float = Field(
        description='Spacing for band distances for automatic kpoints generation, used by seekpath-aiida mode.',
        default=0.025,
    )
    line_density: float = Field(
        description='Density of the point along the path, used by the sumo interface.',
        default=20,
    )
    dos_kpoints_distance: float = Field(
        description=(
            'Kpoints for running DOS calculations in A^-1 * 2pi. Will perform non-SCF DOS calculation is supplied.'
        ),
        default=0.03,
    )
    only_dos: bool = Field(
        description='Flag for running only DOS calculations',
        default=False,
    )
    run_dos: bool = Field(
        description='Flag for running DOS calculations',
        default=False,
    )
    additional_band_analysis_parameters: dict = Field(
        description='Additional keyword arguments for the seekpath/ interface, available keys are:'
        "  ['with_time_reversal', 'reference_distance', 'recipe', 'threshold', 'symprec', 'angle_tolerance']",
        default={},
    )
    kpoints_per_split: int = Field(
        description='Number of kpoints per split for the band structure calculation',
        default=80,
    )
    hybrid_reuse_wavecar: bool = Field(
        description='Whether to reuse the WAVECAR from the previous relax/singlepoint calculation', default=False
    )



"""
Parameter related utils

Contains utils and definitions that are used together with the parameters.
"""
# pylint: disable=too-many-branches

from __future__ import annotations

import enum
from copy import deepcopy
from pathlib import Path
from typing import Any
from warnings import warn

from aiida import orm
from aiida.common.extendeddicts import AttributeDict
from yaml import safe_load

from aiida_vasp.utils.extended_dicts import update_nested_dict

_BASE_NAMESPACES: list[str] = ['electronic', 'smearing', 'charge', 'dynamics', 'bands', 'relax', 'converge']
_DEFAULT_OVERRIDE_NAMESPACE: str = 'incar'


class ChargeEnum(enum.IntEnum):
    """
    Encode values for the initial charge density.

    See: https://www.vasp.at/wiki/index.php/ICHARG
    """

    WAVE = 0
    CHARGE = 1
    ATOMIC = 2
    POTENTIAL = 4
    CONSTANT_CHARGE = 11
    CONSTANT_ATOMIC = 12


class IntSmearingEnum(enum.IntEnum):
    """
    Encode values for the smearing used during integration in reciprocal space.

    See: https://www.vasp.at/wiki/index.php/ISMEAR.
    """

    MP = 1  # pylint: disable=invalid-name
    GAUSSIAN = 0
    FERMI = -1
    PARTIAL = -2
    TETRA = -5


class OrbitEnum(enum.IntEnum):
    """
    Encode values for the projector information.

    See: https://www.vasp.at/wiki/index.php/LORBIT
    """

    ATOM = 0
    ATOM_LM = 1
    ATOM_LM_PHASE = 2
    NO_RWIGS_ATOM = 10
    NO_RWIGS_ATOM_LM = 11
    NO_RWIGS_ATOM_LM_PHASE = 12
    NO_RWIGS_ATOM_LM_PHASE_AUTO = 14
    ATOM_LM_WAVE = 5

    @classmethod
    def get_lorbit_from_combination(cls, **kwargs) -> 'OrbitEnum':
        """Get the correct mode of the projectors/decomposition."""
        combination = tuple(kwargs[i] for i in ['lm', 'phase', 'wigner_seitz_radius'])
        value_from_combinations = {
            (False, False, True): cls.ATOM,
            (True, False, True): cls.ATOM_LM,
            (True, True, True): cls.ATOM_LM_PHASE,
            (False, False, False): cls.NO_RWIGS_ATOM,
            (True, False, False): cls.NO_RWIGS_ATOM_LM,
            (True, True, False): cls.NO_RWIGS_ATOM_LM_PHASE,
            # Not supported, so also calculate lm decomposed
            (False, True, True): cls.ATOM_LM_PHASE,
            # Not supported, so also calculate lm decomposed
            (False, True, False): cls.NO_RWIGS_ATOM_LM_PHASE,
        }
        return value_from_combinations[combination]


class RelaxAlgoEnum(enum.IntEnum):
    """
    Encode values for algorithm descriptively.

    See: https://www.vasp.at/wiki/index.php/IBRION
    """

    NO_UPDATE = -1
    IONIC_RELAXATION_RMM_DIIS = 1
    IONIC_RELAXATION_CG = 2


class RelaxModeEnum(enum.IntEnum):
    """
    Encode values for degrees of freedom mode of relaxation descriptively.

    See: https://cms.mpi.univie.ac.at/wiki/index.php/ISIF
    """

    POS_ONLY = 2
    POS_SHAPE_VOL = 3
    POS_SHAPE = 4
    SHAPE_ONLY = 5
    SHAPE_VOL = 6
    VOL_ONLY = 7

    @classmethod
    def get_isif_from_dof(cls, **kwargs) -> 'RelaxModeEnum':
        """Get the correct mode of relaxation for the given degrees of freedom."""
        relax_choices = ('positions', 'shape', 'volume')  # pylint: disable=invalid-name
        dof = tuple(kwargs[i] for i in relax_choices)
        value_from_dof = {
            (True, False, False): cls.POS_ONLY,
            (True, True, True): cls.POS_SHAPE_VOL,
            (True, True, False): cls.POS_SHAPE,
            (False, True, False): cls.SHAPE_ONLY,
            (False, True, True): cls.SHAPE_VOL,
            (False, False, True): cls.VOL_ONLY,
        }
        try:
            return value_from_dof[dof]
        except KeyError as no_dof:
            raise ValueError(f'Invalid combination for degrees of freedom: {dict(zip(relax_choices, dof))}') from no_dof


class ParametersMassage:  # pylint: disable=too-many-instance-attributes
    """
    A class that contains all relevant massaging of the input parameters for VASP.

    The idea is that this class accepts the set input parameters from AiiDA (non code specifics), checks if any code
    specific parameters supplied are valid VASP input parameters (only rudimentary at this point, should also cross
    check and check types) and convert the AiiDA input parameters to VASP specific parameters. A set function needs to
    be developed for each VASP INCAR parameter that we want to set based on supplied AiiDA/AiiDA-VASP specific
    parameters. These set functions takes these parameters
    and converts it to VASP INCAR compatible tags. The parameter property should return ready to go parameters
    containing the default override namespace, the namespaces set in the `_set_extra_parameters` function and any
    additional namespaces
    that might have been set using the `additional_override_namespaces` setting entry in settings
    that can be supplied to the VaspWorkChain.

    The default override namespace (see `_DEFAULT_OVERRIDE_NAMESPACE`) should always be present when using this VASP
    plugin.
    If using additional plugins, one can for instance supply additional namespace override that can be used,
    depending on what is needed in those plugins and how you construct your workchains.
    """

    _valid_parameters_and_mtime: tuple[list[str], int] = None

    def __init__(
        self,
        parameters: Any,
        unsupported_parameters: Any = None,
        settings: Any = None,
        skip_parameters_validation: bool = False,
    ) -> None:
        self.exit_code = None

        # Flag for skipping any validations
        self._skip_validation = skip_parameters_validation

        # Check type of parameters and set
        self._parameters = check_inputs(parameters)

        # Check type of supplied unsupported_parameters and set
        self._unsupported_parameters = check_inputs(unsupported_parameters)

        # Check type of settings and set
        self._settings = check_inputs(settings)

        # Check setting for any possible supplied additional override namespace and set
        self._additional_override_namespaces = self._fetch_additional_override_namespaces()

        self._massage = AttributeDict()
        # Initialize any allowed override namespaces
        self._massage[_DEFAULT_OVERRIDE_NAMESPACE] = AttributeDict()
        for item in self._additional_override_namespaces:
            self._massage[item] = AttributeDict()

        self._check_valid_namespaces()
        # Load the valid INCAR parameters that are supported by VASP
        self._load_valid_params()
        # Establish set functions which will convert AiiDA/AiiDA-VASP specific parameters to VASP INCAR tags
        self._functions = ParameterSetFunctions(self._parameters, self._massage[_DEFAULT_OVERRIDE_NAMESPACE])
        # Convert general parameters to VASP specific ones, using the set functions
        self._set_vasp_parameters()
        # Override or set parameters that are supplied in the default override namespace
        # (should be valid VASP INCAR tags)
        self._set_override_vasp_parameters()
        # Set any extra parameters not related to INCAR
        self._set_extra_vasp_parameters()
        # Set any additional override namespace that should just be forwarded
        self._set_additional_override_parameters()
        # Finally, we validate the INCAR parameters in order to prepare it for dispatch
        self._validate_vasp_parameters()

    def _check_valid_namespaces(self) -> None:
        """Check that we do not have namespaces on the input parameters that is unsupported."""
        for key in self._parameters.keys():
            if key not in list(_BASE_NAMESPACES + self._additional_override_namespaces + [_DEFAULT_OVERRIDE_NAMESPACE]):
                raise ValueError(f'The supplied namespace: {key} is not supported.')

    def _load_valid_params(self) -> None:
        """Import a list of valid parameters for VASP. This is generated from the manual."""

        # Cache the valid parameters as class attributes to void repeated loading
        filepath = Path(__file__).parent / 'parameters.yml'
        mtime = Path(filepath).stat().st_mtime

        if (
            ParametersMassage._valid_parameters_and_mtime is None
            or mtime != ParametersMassage._valid_parameters_and_mtime[1]
        ):
            with open(filepath, 'r', encoding='utf8') as handler:
                tags_data = safe_load(handler)
                ParametersMassage._valid_parameters_and_mtime = (list(tags_data.keys()), mtime)

        # Now add any unsupported parameter to the list
        self._valid_parameters = deepcopy(ParametersMassage._valid_parameters_and_mtime[0])
        for key, _ in self._unsupported_parameters.items():
            if key.lower() not in self._valid_parameters:
                self._valid_parameters.append(key.lower())

    def _fetch_additional_override_namespaces(self) -> list[str]:
        """Check the settings for any additional supplied override namespace and return it."""
        try:
            override_namespaces = self._settings.additional_override_namespaces
        except AttributeError:
            override_namespaces = []

        return override_namespaces

    def _set_vasp_parameters(self) -> None:
        """Iterate over the valid parameters and call the set function associated with that parameter."""
        for key in self._valid_parameters:
            self._set(key)

    def _set_override_vasp_parameters(self) -> None:
        """Set the any supplied override parameters."""
        if _DEFAULT_OVERRIDE_NAMESPACE not in self._parameters:
            return

        for key, item in self._parameters[_DEFAULT_OVERRIDE_NAMESPACE].items():
            # Sweep the override input parameters (only care about the ones in the default override namespace)
            # to check if they are valid VASP tags.
            if self._valid_vasp_parameter(key.lower()):
                # Add or override in the default override namespace
                self._massage[_DEFAULT_OVERRIDE_NAMESPACE][key.lower()] = item
            else:
                raise ValueError(f'The supplied key: {key} is not a support VASP parameter.')

    def _set_extra_vasp_parameters(self) -> None:
        """
        Find if there are any extra parameters that are not part of the INCAR that needs to be set.

        One example is the dynamic namespace which handles for instance flags for selective dynamics.
        These flags are more connected to a calculation than a StructureData and thus it was necessary
        to make sure it was valid input to the VASP workchain.

        """
        # return if dynamics is not supplied
        if 'dynamics' not in self._parameters:
            return

        if self._parameters.dynamics:
            self._massage.dynamics = AttributeDict()

        for key_, item in self._parameters.dynamics.items():
            key = key_.lower()
            if key in ['positions_dof']:
                self._massage.dynamics[key] = item
            else:
                warn(f"Key {key} is not supported for 'dynamics' input.")

    def _set_additional_override_parameters(self) -> None:
        """Set any customized parameter namespace, including its content on the massaged container."""
        parameters_keys = self._parameters.keys()
        for item in self._additional_override_namespaces:
            if item in parameters_keys:
                # Only add if namespace exists in parameters
                self._massage[item] = AttributeDict(self._parameters[item])

    def _valid_vasp_parameter(self, key: str) -> bool:
        """Make sure a key are recognized as a valid VASP input parameter."""
        if self._skip_validation or (key in self._valid_parameters):
            return True

        return False

    def _validate_vasp_parameters(self) -> None:
        """Make sure all the massaged values are recognized as valid VASP input parameters."""

        for key in self._massage[_DEFAULT_OVERRIDE_NAMESPACE]:
            if not self._valid_vasp_parameter(key.lower()):
                raise ValueError(f'The supplied key: {key} is not a support VASP parameter.')

    def _set(self, key: str) -> None:
        """Call the necessary function to set each parameter."""
        try:
            getattr(self._functions, 'set_' + key)()
        except AttributeError:
            # We have no setter function for the valid key, meaning there is no general parameter that is linked
            # to this key (INCAR tag). These tags have to be supplied in the default override namespace for now.
            pass

    @property
    def parameters(self) -> AttributeDict:
        """Return the massaged parameter set ready to go in VASP format."""
        return self._massage


class ParameterSetFunctions:
    """Container for the set functions that converts an AiiDA parameters to a default override specific one."""

    def __init__(self, parameters: AttributeDict, incar: AttributeDict) -> None:
        self._parameters = parameters
        self._incar = incar

    def set_encut(self) -> None:
        """
        Set which plane wave cutoff to use.

        See https://www.vasp.at/wiki/index.php/ENCUT
        """

        try:
            self._incar.encut = self._parameters.electronic.pwcutoff
        except AttributeError:
            pass

    def set_ibrion(self) -> None:
        """
        Set which algorithm to use for ionic movements.

        See: https://www.vasp.at/wiki/index.php/IBRION
        """

        if self._relax():
            try:
                if self._parameters.relax.algo == 'cg':
                    self._incar.ibrion = RelaxAlgoEnum.IONIC_RELAXATION_CG.value
                elif self._parameters.relax.algo == 'rd':
                    self._incar.ibrion = RelaxAlgoEnum.IONIC_RELAXATION_RMM_DIIS.value
                else:
                    raise ValueError(f'The supplied relax.algo: {self._parameters.relax.algo} is not supported')
            except AttributeError:
                pass

    def set_ediffg(self) -> None:
        """
        Set the cutoff to use for relaxation.

        See: https://www.vasp.at/wiki/index.php/EDIFFG
        """

        if not self._relax():
            # This flag is only valid if you have enabled relaxation
            return
        energy_cutoff = False
        try:
            if self._parameters.relax.energy_cutoff is not None:
                self._incar.ediffg = self._parameters.relax.energy_cutoff
                energy_cutoff = True
        except AttributeError:
            pass
        try:
            self._incar.ediffg = -abs(self._parameters.relax.force_cutoff)
            if energy_cutoff:
                raise ValueError('User supplied both a force and an energy cutoff for the relaxation. Please select.')
        except AttributeError:
            pass

    def set_nsw(self) -> None:
        """
        Set the number of ionic steps to perform.

        See: https://www.vasp.at/wiki/index.php/NSW
        """

        if self._relax():
            try:
                self._set_simple('nsw', self._parameters.relax.steps)
            except AttributeError:
                pass

    def set_isif(self) -> None:
        """
        Set relaxation mode according to the chosen degrees of freedom.

        See: https://www.vasp.at/wiki/index.php/ISIF
        """

        if self._relax():
            positions = self._parameters.get('relax', {}).get('positions', False)
            shape = self._parameters.get('relax', {}).get('shape', False)
            volume = self._parameters.get('relax', {}).get('volume', False)
            try:
                self._incar.isif = RelaxModeEnum.get_isif_from_dof(
                    positions=positions, shape=shape, volume=volume
                ).value
            except AttributeError:
                pass

    def set_ismear(self) -> None:
        """
        Make sure we do not supply invalid integration methods when running explicit k-point grids.

        See: https://www.vasp.at/wiki/index.php/ISMEAR
        """

        try:
            if self._parameters.smearing.gaussian:
                self._set_simple('ismear', IntSmearingEnum.GAUSSIAN.value)
        except AttributeError:
            pass
        try:
            if self._parameters.smearing.fermi:
                self._set_simple('ismear', IntSmearingEnum.FERMI.value)
        except AttributeError:
            pass
        try:
            if self._parameters.smearing.mp:
                self._set_simple('ismear', IntSmearingEnum.MP.value * abs(int(self._parameters.smearing.mp)))
        except AttributeError:
            pass
        try:
            if self._parameters.smearing.tetra:
                self._set_simple('ismear', IntSmearingEnum.TETRA.value)
        except AttributeError:
            pass

    def set_icharg(self) -> None:
        """
        Set the flag to start from input charge density and keep it constant.

        See: https://www.vasp.at/wiki/index.php/ICHARG
        """

        try:
            if self._parameters.charge.from_wave:
                self._set_simple('icharg', ChargeEnum.WAVE.value)
        except AttributeError:
            pass
        try:
            if self._parameters.charge.from_charge:
                self._set_simple('icharg', ChargeEnum.CHARGE.value)
        except AttributeError:
            pass
        try:
            if self._parameters.charge.from_atomic:
                self._set_simple('icharg', ChargeEnum.ATOMIC.value)
        except AttributeError:
            pass
        try:
            if self._parameters.charge.from_potential:
                self._set_simple('icharg', ChargeEnum.POTENTIAL.value)
        except AttributeError:
            pass
        try:
            if self._parameters.charge.constant_charge:
                self._set_simple('icharg', ChargeEnum.CONSTANT_CHARGE.value)
        except AttributeError:
            pass
        try:
            if self._parameters.charge.constant_atomic:
                self._set_simple('icharg', ChargeEnum.CONSTANT_ATOMIC.value)
        except AttributeError:
            pass

    def set_lorbit(self) -> None:
        """
        Set the flag that controls the projectors/decomposition onto orbitals.

        See: https://www.vasp.at/wiki/index.php/LORBIT
        """
        self._set_wigner_seitz_radius()
        try:
            if self._parameters.bands.decompose_bands:
                if self._parameters.bands.decompose_wave:
                    # Issue a warning that one can only use either or
                    raise ValueError('Only projections/decompositions on the bands or the wave function are allowed.')
                wigner_seitz_radius = False
                try:
                    if abs(self._incar.rwigs[0]) > 1e-8:
                        wigner_seitz_radius = True
                except AttributeError:
                    pass
                if self._parameters.bands.decompose_auto:
                    self._set_simple('lorbit', OrbitEnum.NO_RWIGS_ATOM_LM_PHASE_AUTO.value)
                else:
                    try:
                        lm = self._parameters.bands.lm  # pylint: disable=invalid-name
                    except AttributeError:
                        lm = False  # pylint: disable=invalid-name
                    try:
                        phase = self._parameters.bands.phase
                    except AttributeError:
                        phase = False
                    lorbit = OrbitEnum.get_lorbit_from_combination(
                        lm=lm, phase=phase, wigner_seitz_radius=wigner_seitz_radius
                    ).value
                    self._set_simple('lorbit', lorbit)
            else:
                try:
                    if self._parameters.bands.decompose_wave:
                        self._set_simple('lorbit', OrbitEnum.ATOM_LM_WAVE.value)
                except AttributeError:
                    pass
        except AttributeError:
            try:
                if self._parameters.bands.decompose_wave:
                    self._set_simple('lorbit', OrbitEnum.ATOM_LM_WAVE.value)
            except AttributeError:
                pass

    def _set_wigner_seitz_radius(self) -> None:
        """
        Set the Wigner Seitz radius that is used to project/decompose.

        See: https://www.vasp.at/wiki/index.php/RWIGS
        """
        try:
            wigner_seitz_radius = self._parameters.bands.wigner_seitz_radius
            # Check that it is defined as a list
            if isinstance(wigner_seitz_radius, list):
                if wigner_seitz_radius[0]:
                    self._set_simple('rwigs', wigner_seitz_radius)
            else:
                raise ValueError(
                    'The parameter wigner_seitz_radius should be supplied as a list of floats bigger than zero.'
                )
        except AttributeError:
            pass

    def _relax(self) -> bool:
        """Check if we have enabled relaxation."""
        return bool(
            self._parameters.get('relax', {}).get('positions')
            or self._parameters.get('relax', {}).get('shape')
            or self._parameters.get('relax', {}).get('volume')
        )

    def _set_simple(self, target: str, value: Any) -> None:
        """Set basic parameter."""
        try:
            self._incar[target] = value
        except AttributeError:
            pass


def check_inputs(supplied_inputs: None | AttributeDict | orm.Dict | dict) -> AttributeDict:
    """Check that the inputs are of some correct type and returned as AttributeDict."""
    inputs = None
    if supplied_inputs is None:
        inputs = AttributeDict()
    elif isinstance(supplied_inputs, orm.Dict):
        inputs = AttributeDict(supplied_inputs.get_dict())
    elif isinstance(supplied_inputs, dict):
        inputs = AttributeDict(supplied_inputs)
    elif isinstance(supplied_inputs, AttributeDict):
        inputs = supplied_inputs
    else:
        raise ValueError(
            f'The supplied type {type(inputs)} of inputs is not supported. Supply a dict, Dict or an AttributeDict.'
        )

    return inputs


def inherit_and_merge_parameters(inputs: dict[str, Any]) -> AttributeDict:
    """
    Goes trough the inputs namespaces and the namespaces in the inputs.parameters and merge them.

    Note that parameters specified in the inputs.parameters will override what is supplied as workchain input,
    in case there is overlap.
    """
    parameters = AttributeDict()
    namespaces = _BASE_NAMESPACES

    # We start with a clean parameters and first set the allowed namespaces and its content from the inputs of the
    # workchain
    for namespace in namespaces:  # pylint: disable=too-many-nested-blocks
        parameters[namespace] = AttributeDict()
        try:
            for key, item in inputs[namespace].items():
                if isinstance(item, orm.ArrayData):
                    # Only allow one array per input
                    if len(item.get_arraynames()) > 1:
                        raise IndexError(
                            f'The input array with a key {key} contains more than one array. '
                            'Please make sure an input only contains one array.'
                        )
                    for array in item.get_arraynames():
                        parameters[namespace][key] = item.get_array(array)
                elif isinstance(item, orm.Dict):
                    parameters[namespace][key] = item.get_dict()
                elif isinstance(item, orm.List):
                    parameters[namespace][key] = item.get_list()
                else:
                    parameters[namespace][key] = item.value
        except KeyError:
            pass

    # Then obtain the inputs.parameters.
    # Here we do not do any checks for valid parameters, that is done later when reaching the ParameterMassager.
    try:
        input_parameters = AttributeDict(inputs.parameters.get_dict())
    except AttributeError:
        # Inputs might not have parameters
        input_parameters = AttributeDict()

    # Now the namespace and content of the workchain inputs and the inputs.parameters are merged.
    # Any supplied namespace in the parameters (i.e. inputs.parameters.somekey) will override what
    # is supplied to the workchain input namespace (i.e. inputs.somekey).
    # We cannot use regular update here, as we only want to replace each key if it exists, if a key
    # contains a new dict we need to traverse that, hence we have a function to perform this update.
    update_nested_dict(parameters, input_parameters)

    return parameters


# class ParserSettingsChecker:
#     """
#     Check for inconsistency between the parser settings and the INCAR tags.
#     """

#     def __init__(self, parameters: dict, settings: dict, default_settings=None):
#         """
#         Instantiate an ParserSettingsChecker object.

#         :param parameters: A dictionary of INCAR tags to be used
#         :param settings: The settings dictionary
#         :param default_option: The default options to be used for instantiating the ParserSetting object.

#         """

#         if default_settings is None:
#             default_settings = DEFAULT_SETTINGS

#         # Obtain the expected quantities names to be parsed
#         self.settings = ParserSettings(settings, default_settings)
#         self.quantities = settings.quantity_names_to_parse
#         self.parameters = parameters

#     def check(self):
#         """
#         Check the consistency between the supplied INCAR tags the requested quantities to be parsed.
#         """

#         self.check_maximum_stress()
#         self.check_wavecar_chgcar()
#         self.check_born_charges()
#         self.check_dynmat()

#     def check_maximum_stress(self):
#         """Check the the maximum_stress can be parsed"""

#         if 'maximum_stress' not in self.quantities and 'stress' not in self.quantities:
#             return

#         isif = self.parameters.get('isif')
#         ibrion = self.parameters.get('ibrion', -1)
#         lhfcalc = self.parameters.get('lhfcalc', False)
#         if isif is None:
#             if ibrion == 0:
#                 isif = 0
#             if lhfcalc:
#                 isif = 0

#         if isif == 0:
#             raise InputValidationError(
#                 'Requested to parse <maximum_stress> but it would not be calculated due to ISIF settings.'
#             )

#     def check_wavecar_chgcar(self):
#         """Check if WAVECAR CHGCAR are set to be written"""

#         if 'wavecar' in self.quantities and not self.parameters.get('lwave', True):
#             raise InputValidationError('Requested to retrieve <WAVECAR> but it not set to be written.')

#         if 'chgcar' in self.quantities and not self.parameters.get('lwave', True):
#             raise InputValidationError('Requested to retrieve <CHGCAR> but it not set to be written.')

#     def check_born_charges(self):
#         """Check if projectors can be parsed"""
#         if 'born_charges' not in self.quantities and 'dielectrics' not in self.quantities:
#             return

#         lepsilon = self.parameters.get('lepsilon', False)
#         if not lepsilon:
#             raise InputValidationError(
#                 'Requested to parse "born_charges"/"dielectrics" but they are not going to be calculated.'
#             )

#     def check_dynmat(self):
#         if 'hessian' not in self.quantities and 'dynmat' not in self.quantities:
#             return
#         ibrion = self.parameters.get('ibrion', -1)
#         if ibrion not in [5, 6, 7, 8]:
#             raise InputValidationError('Requsted to parse "hessian"/"dynmat" '
#                                      'but they are not going to be calculated.')



from __future__ import annotations

import gzip
import os
import re
import shutil
from pathlib import Path
from typing import Any

from aiida import orm
from aiida.common.links import LinkType
from aiida.orm import CalcJobNode, Node, QueryBuilder, WorkChainNode
from aiida.plugins import WorkflowFactory
from aiida.repository import FileType

from aiida_vasp.parsers.content_parsers.poscar import PoscarParser
from aiida_vasp.parsers.content_parsers.potcar import MultiPotcarIo
from aiida_vasp.workchains.v2.relax import VaspRelaxWorkChain

from .aiida_utils import ensure_node_first_arg, ensure_node_kwargs


def export_vasp(
    process_node: orm.Node, dst: str | Path, decompress: bool = False, include_potcar: bool = False
) -> None:
    """Export VASP calculation files to destination."""

    # Dispatch export function based on process type
    if process_node.process_type.endswith('vasp'):
        _export_calculation(process_node, dst, decompress=decompress, include_potcar=include_potcar)
    elif process_node.process_type.endswith('neb'):
        export_neb(process_node, dst, decompress=decompress, include_potcar=include_potcar)
    elif process_node.process_type.endswith('relax'):
        _export_workchain(process_node, dst, decompress=decompress, include_potcar=include_potcar)
    else:
        raise TypeError(f'Unsupported process type: {process_node.process_type}')


@ensure_node_kwargs
@ensure_node_first_arg
def _export_calculation(
    calc_node: orm.CalcJobNode, dst: Path, decompress: bool = False, include_potcar: bool = False
) -> None:
    """Export a single calculation."""

    folder = Path(dst)
    folder.mkdir(exist_ok=True)

    # Inputs
    retrieved = calc_node.get_outgoing(link_label_filter='retrieved').one().node
    if isinstance(calc_node, CalcJobNode):
        calcjob = calc_node
    elif isinstance(calc_node, WorkChainNode):
        # In this case the node is an workchain we export the
        # 'retrieved' output link and trace to its ancestor
        calcjob = retrieved.base.links.get_incoming(link_label_filter='retrieved', link_type=LinkType.CREATE).one().node
    else:
        raise RuntimeError(f'The node {calc_node} is not a valid calculation')
    info_file = folder / ('aiida_info')
    info_content = f'Label: {calcjob.label}\nDescription: {calcjob.description}\nUUID: {calcjob.uuid}\n'
    info_file.write_text(info_content)
    # export the retrieved outputs  and the input files
    save_all_repository_objects(retrieved, folder, decompress)
    save_all_repository_objects(calcjob, folder, decompress)
    if include_potcar:
        export_pseudos(calcjob, folder)


@ensure_node_first_arg
@ensure_node_kwargs
def export_pseudos(calc_job_node: Any, folder: Path) -> None:
    """Save the pseudopotential file (POTCAR)"""
    pps = calc_job_node.get_incoming(link_label_filter='potential%').nested()['potential']
    multi_potcar = MultiPotcarIo.from_structure(calc_job_node.inputs.structure, pps)
    dst = str(folder / 'POTCAR')
    multi_potcar.write(dst)


@ensure_node_first_arg
@ensure_node_kwargs
def _export_workchain(
    work_node: orm.WorkChainNode, dst: Path, decompress: bool = False, include_potcar: bool = False
) -> None:
    """Export a workchain."""

    dst = Path(dst)
    dst.mkdir(exist_ok=True)
    if work_node.process_class not in (VaspRelaxWorkChain):
        raise ValueError(
            f'Error {work_node} should be `VaspRelaxWorkChain` or `RelaxWorkChain`, but it is {work_node.process_class}'
        )

    q = QueryBuilder()
    q.append(Node, filters={'id': work_node.pk})
    q.append(WorkChainNode, tag='vaspwork', project=['id', '*'])
    q.order_by({'vaspwork': {'id': 'asc'}})  # Sort by ascending PK
    for index, (pk, node) in enumerate(q.iterall()):
        relax_folder = dst / f'relax_calc_{index:03d}'
        try:
            _export_calculation(node, relax_folder, decompress=decompress, include_potcar=include_potcar)
        except (ValueError, AttributeError, KeyError):
            print(f'Error exporting calculation {pk}')

    # Write POSCAR file for the input
    input_structure = work_node.inputs.structure
    poscar_parser = PoscarParser(data=input_structure, precision=10)
    poscar_parser.write(str(dst / 'POSCAR'))

    # Write POSCAR file for the input
    try:
        out_structure = work_node.outputs.relax.structure
    except AttributeError:
        print(
            'Cannot find the output structure - skipping.'
            ' This usually means that the relaxation did not finish without error.'
        )
        out_structure = None
    if out_structure:
        poscar_parser = PoscarParser(data=out_structure, precision=10)
        poscar_parser.write(str(dst / 'POSCAR_RELAXED'))

    # Write the info
    info_file = dst / ('aiida_info')
    info_content = f'Label: {work_node.label}\nDescription: {work_node.description}\nUUID: {work_node.uuid}\n'
    info_file.write_text(info_content)


@ensure_node_first_arg
def export_neb(
    workchain: Any,
    dst: str | Path,
    decompress: bool = True,
    include_potcar: bool = True,
    energy_type: str = 'energy_extrapolated',
) -> None:
    """Export the neb calculation"""
    energies = {key: value[energy_type] for key, value in workchain.outputs.misc['total_energies'].items()}

    # Query for the energy computed for the end structures
    q = orm.QueryBuilder()
    q.append(orm.Node, filters={'id': workchain.inputs.initial_structure.id}, tag='root')
    q.append(orm.CalcFunctionNode, with_outgoing='root', project=['attributes.function_name'])
    q.append(
        orm.StructureData,
        with_outgoing=orm.CalcFunctionNode,
        tag='relaxed',
        project=['label'],
        # edge_filters={'label': 'init_structure'},
        edge_project=['label'],
    )
    q.append(
        WorkflowFactory('vasp.v2.relax'),
        with_outgoing='relaxed',
        project=['label', 'uuid'],
        tag='relaxation',
    )
    q.append(
        orm.Dict,
        with_incoming='relaxation',
        edge_filters={'label': 'misc'},
        project=['attributes.total_energies.energy_extrapolated'],
    )
    q.append(orm.CalcJobNode, with_outgoing=orm.Dict, project=['*'])
    q.distinct()

    # First export the original calculation
    _export_calculation(workchain, dst, decompress=decompress, include_potcar=include_potcar)
    ends = {}
    end_id = f'{len(energies) + 1:02d}'
    for _, _, _, relax_uuid, eng, calcjob, label in q.all():
        if label.startswith('init'):
            if '00' in ends:
                print(
                    'Duplicated calculation: {relax_uuid} -> {eng} vs existing {existing}'.format(
                        relax_uuid=relax_uuid, eng=eng, existing=ends['00']
                    )
                )
            else:
                ends['00'] = calcjob

        elif label.startswith('final'):
            if end_id in ends:
                print(
                    'Duplicated calculation: {relax_uuid} -> {eng} vs existing {existing}'.format(
                        relax_uuid=relax_uuid, eng=eng, existing=ends[end_id]
                    )
                )
            else:
                ends[end_id] = calcjob
    # Export the end point calculation
    for key, value in ends.items():
        _export_calculation(value, Path(dst) / key, decompress=decompress, include_potcar=include_potcar)


@ensure_node_kwargs
def copy_from_aiida(name: str, node: orm.Node, dst: Path, decompress: bool = False, exclude: str | None = None) -> None:
    """Copy objects from aiida repository.

    :param name: The full name (including the parent path) of the object.
    :type name: str
    :param node: Node object for which the files in the repo to be copied.
    :type node: orm.Node
    :param dst: Path of the destination folder.
    :type dst: Path
    :param decompress: Whether to decompress files
    :type decompress: bool, optional
    :param exclude: Pattern to exclude files
    :type exclude: str, optional

    This is a recursive function so directory copying also works.
    """

    # For check the regex the first thing because this function will be called recursively
    if exclude and re.match(exclude, name):
        return

    obj = node.base.repository.get_object(name)

    # If it is a directory, copy the contents one by one
    if obj.file_type == FileType.DIRECTORY:
        for sub_obj in node.base.repository.list_objects(name):
            copy_from_aiida(os.path.join(name, sub_obj.name), node, dst, exclude=exclude)
    else:
        # It is a file
        with node.base.repository.open(name, mode='rb') as fsource:
            # Make parent directory if needed
            frepo_path = dst / name
            Path(frepo_path.parent).mkdir(exist_ok=True, parents=True)
            # Write the file
            if name.endswith('.gz') and decompress:
                out_path = str(frepo_path)[:-3]
                out_decompress = True
            else:
                out_decompress = False
                out_path = str(frepo_path)

            if not out_decompress:
                with open(out_path, 'wb') as fdst:
                    shutil.copyfileobj(fsource, fdst)
            else:
                gobj = gzip.GzipFile(fileobj=fsource, mode='rb')
                with open(out_path, 'wb') as fdst:
                    shutil.copyfileobj(gobj, fdst)


@ensure_node_first_arg
def save_all_repository_objects(
    node: orm.Node, target_path: Path, decompress: bool = False, exclude: str | None = None
) -> None:
    """Copy all objects of a node saved in the repository to the disc"""
    for name in node.list_object_names():
        copy_from_aiida(name, node, target_path, decompress, exclude=exclude)



"""
Bands workchain with a more flexible input
TODO:

- Add option to use alternative pathways obtained using sumo-interface
- Improve the hybrid workchain by performing local dryrun to extract the full kpoints
    - If running SOC, the ISYM should be turned to 0 or -1.
"""

from copy import deepcopy
from logging import getLogger
from typing import Any, List

import numpy as np
from aiida import orm
from aiida.common.extendeddicts import AttributeDict
from aiida.common.links import LinkType
from aiida.engine import WorkChain, append_, calcfunction, if_
from aiida.orm.nodes.data.base import to_aiida_type
from aiida.plugins import WorkflowFactory
from aiida.tools import get_explicit_kpoints_path
from ase.geometry import cell_to_cellpar

from aiida_vasp.data.chargedensity import ChargedensityData
from aiida_vasp.parsers.content_parsers.vasprun import VasprunParser
from aiida_vasp.utils.extended_dicts import update_nested_dict, update_nested_dict_node
from aiida_vasp.utils.kmesh import get_ir_kpoints_data
from aiida_vasp.utils.opthold import BandOptions

from .common import OVERRIDE_NAMESPACE
from .common.dryrun import dryrun_relax_builder
from .common.transform import magnetic_structure_decorate, magnetic_structure_dedecorate
from .mixins import WithBuilderUpdater

SITE_MAG_THRESHOLD = 0  # Threshold for considering a site to be magnetic


logger = getLogger(__name__)


class VaspBandsWorkChain(WorkChain, WithBuilderUpdater):
    """
    Workchain for running bands calculations.

    This workchain does the following:

    1. Relax the structure if requested (eg. inputs passed to the relax namespace).
    2. Do a SCF singlepoint calculation.
    3. Do a non-scf calculation for bands and dos.

    Inputs must be passed for the SCF calculation, others are optional. The dos calculation
    will only run if the kpoints for DOS are passed or a full specification is given under the
    `dos` input namesace.

    The SCF calculation may be skipped by passing a CHGCAR file/remote folder. In which case the SCF inputs
    are carried on for non-scf calculations.

    The band structure calculation will run unless `only_dos` is set to `Bool(True)`.

    For magnetic structures, the workchain will internally create additional species for the symmetry
    analysis and regenerate "undecorated" structures with corresponding initial magnetic moments. This
    works for both FM and AFM species. Care should be taken when the MAGMOM is obtained from site projected
    values in case of unexpected symmetry breaking.

    Input for bands and dos calculations are optional. However, if they are needed, the full list of inputs must
    be passed. For the `parameters` node, one may choose to only specify those fields that need to be updated.

    For optics calculations, one should run with `only_dos`, set 'NBANDS' to a high value and
    set 'LOPTICS' to be True.
    """

    _base_wk_string = 'vasp.v2.vasp'
    _relax_wk_string = 'vasp.v2.relax'
    option_class = BandOptions

    @classmethod
    def define(cls, spec: Any) -> None:
        """Initialise the WorkChain class"""
        super().define(spec)
        relax_work = WorkflowFactory(cls._relax_wk_string)
        base_work = WorkflowFactory(cls._base_wk_string)

        spec.input('structure', help='The input structure', valid_type=orm.StructureData)
        spec.input(
            'bs_kpoints',
            help='Explicit kpoints for the bands. Will not generate kpoints if supplied.',
            valid_type=orm.KpointsData,
            required=False,
        )
        spec.input(
            'band_settings',
            help=BandOptions.aiida_description(),
            valid_type=orm.Dict,
            validator=BandOptions.aiida_validate,
            serializer=BandOptions.aiida_serialize,
        )
        spec.expose_inputs(
            relax_work,
            namespace='relax',
            exclude=('structure',),
            namespace_options={
                'required': False,
                'populate_defaults': False,
                'help': 'Inputs for Relaxation workchain, if needed',
            },
        )
        spec.expose_inputs(
            base_work,
            namespace='scf',
            exclude=('structure',),
            namespace_options={
                'required': True,
                'populate_defaults': True,
                'help': 'Inputs for SCF workchain, mandatory',
            },
        )
        spec.expose_inputs(
            base_work,
            namespace='bands',
            exclude=('structure', 'kpoints'),
            namespace_options={
                'required': False,
                'populate_defaults': False,
                'help': 'Inputs for bands calculation, if needed',
            },
        )
        spec.expose_inputs(
            base_work,
            namespace='dos',
            exclude=('structure',),
            namespace_options={
                'required': False,
                'populate_defaults': False,
                'help': 'Inputs for DOS calculation, if needed',
            },
        )
        spec.input(
            'clean_children_workdir',
            valid_type=orm.Str,
            serializer=to_aiida_type,
            help='What part of the called children to clean',
            required=False,
            default=lambda: orm.Str('none'),
        )
        spec.input(
            'chgcar',
            required=False,
            valid_type=ChargedensityData,
            help='Explicit CHGCAR file used for DOS/Bands calculations',
        )
        spec.input(
            'restart_folder',
            required=False,
            valid_type=orm.RemoteData,
            help='A remote folder containing the CHGCAR file to be used',
        )
        spec.outline(
            cls.setup,
            if_(cls.should_do_relax)(
                cls.run_relax,
                cls.verify_relax,
            ),
            if_(cls.should_generate_path)(cls.generate_path),
            if_(cls.should_run_scf)(
                cls.run_scf,
                cls.verify_scf,
            ),
            cls.run_bands_dos,
            cls.inspect_bands_dos,
        )

        spec.output(
            'primitive_structure',
            required=False,
            help='Primitive structure used for band structure calculations',
        )
        spec.output('band_structure', required=False, help='Computed band structure with labels')
        spec.output('seekpath_parameters', help='Parameters used by seekpath', required=False)
        spec.output('dos', required=False)
        spec.output('projectors', required=False)

        spec.exit_code(501, 'ERROR_SUB_PROC_RELAX_FAILED', message='Relaxation workchain failed')
        spec.exit_code(502, 'ERROR_SUB_PROC_SCF_FAILED', message='SCF workchain failed')
        spec.exit_code(
            503,
            'ERROR_SUB_PROC_BANDS_FAILED',
            message='Band structure workchain failed',
        )
        spec.exit_code(504, 'ERROR_SUB_PROC_DOS_FAILED', message='DOS workchain failed')
        spec.exit_code(
            601,
            'ERROR_INPUT_STRUCTURE_NOT_PRIMITIVE',
            message='The input structure is not the primitive one!',
        )

    def select_chgcar_from_inputs(self) -> None:
        """Setup CHGCAR from inputs"""
        if self.inputs.get('chgcar'):
            self.ctx.chgcar = self.inputs.chgcar
            self.report(f'Using CHGCAR {self.inputs.chgcar} from input')
        else:
            self.ctx.chgcar = None

        if self.inputs.get('restart_folder'):
            self.ctx.restart_folder = self.inputs.restart_folder
            self.report(f'Using remote folder {self.inputs.restart_folder} for restart')
        else:
            self.ctx.restart_folder = None

    def setup(self) -> None:
        """Setup the calculation"""
        self.ctx.current_structure = self.inputs.structure
        self.ctx.bs_kpoints = self.inputs.get('bs_kpoints')
        param = self.inputs.scf.parameters.get_dict()
        if 'magmom' in param[OVERRIDE_NAMESPACE] and not self.inputs.band_settings['only_dos']:
            self.report('Magnetic system passed for BS')
            self.ctx.magmom = param[OVERRIDE_NAMESPACE]['magmom']
        else:
            self.ctx.magmom = None

    def should_do_relax(self) -> bool:
        """Wether we should do relax or not"""
        return 'relax' in self.inputs

    def run_relax(self) -> Any:
        """Run the relaxation"""
        relax_work = WorkflowFactory(self._relax_wk_string)
        inputs = self.exposed_inputs(relax_work, 'relax', agglomerate=True)
        inputs = AttributeDict(inputs)
        inputs.metadata.call_link_label = 'relax'
        inputs.structure = self.ctx.current_structure

        # Ensure the WAVECAR is written by the calculation
        if self.inputs.band_settings.get('hybrid_reuse_wavecar', False):
            pdict = inputs.vasp.parameters.get_dict()
            # Update the relax settings so we do not clean the final singepoint calculation
            rdict = inputs.relax_settings.get_dict()
            rdict['keep_sp_workdir'] = True
            if rdict != inputs.relax_settings.get_dict():
                inputs.relax_settings = orm.Dict(dict=rdict)

            pdict['incar']['lwave'] = True
            if pdict != inputs.vasp.parameters.get_dict():
                inputs.vasp.parameters = orm.Dict(dict=pdict)

        running = self.submit(relax_work, **inputs)
        return self.to_context(workchain_relax=running)

    def verify_relax(self) -> Any:
        """Verify the relaxation"""
        relax_workchain = self.ctx.workchain_relax
        if not relax_workchain.is_finished_ok:
            self.report('Relaxation finished with Error')
            return self.exit_codes.ERROR_SUB_PROC_RELAX_FAILED

        # Use the relaxed structure as the current structure
        self.ctx.current_structure = relax_workchain.outputs.relax__structure

    def should_run_scf(self) -> bool:
        """Wether we should run SCF calculation"""
        # Setup the CHGCAR and remote folder input if necessary
        self.select_chgcar_from_inputs()
        # Only need to run SCF calculation when no explicity CHGCAR or folder set
        return not (self.ctx.chgcar or self.ctx.restart_folder)

    def should_generate_path(self) -> bool:
        """
        Seekpath should only run if no explicit bands is provided or we are just
        running for DOS, in which case the original structure is used.
        """
        return 'bs_kpoints' not in self.inputs and (not self.inputs.band_settings['only_dos'])

    def generate_path(self) -> None:
        """
        Run seekpath to obtain the primitive structure and bands
        """

        current_structure_backup = self.ctx.current_structure

        mode = self.inputs.band_settings['band_mode']

        if mode == 'seekpath-aiida':
            inputs = {
                'band_settings': orm.Dict(
                    {
                        'reference_distance': self.inputs.band_settings['band_kpoints_distance'],
                        'symprec': self.inputs.band_settings['symprec'],
                        **self.inputs.band_settings['additional_band_analysis_parameters'],
                    }
                ),
                'metadata': {'call_link_label': 'seekpath'},
            }
            func = seekpath_structure_analysis
        else:
            # Using sumo interface
            try:
                from .common.sumo_kpath import kpath_from_sumo_v2  # noqa: PLC0415
            except ImportError:
                raise ImportError('Sumo is not installed, please install it to use this feature.')

            inputs = {
                'band_settings': orm.Dict(
                    {
                        'line_density': self.inputs.band_settings['line_density'],
                        'symprec': self.inputs.band_settings['symprec'],
                        'mode': mode,
                        **self.inputs.band_settings['additional_band_analysis_parameters'],
                    }
                ),
                'metadata': {'call_link_label': 'sumo_kpath'},
            }
            func = kpath_from_sumo_v2

        magmom = self.ctx.get('magmom', None)

        # For magnetic structures, create different kinds for the analysis in case that the
        # symmetry should be lowered. This also makes sure that the magnetic moments are consistent
        if magmom:
            decorate_result = magnetic_structure_decorate(self.ctx.current_structure, orm.List(list=magmom))
            decorated = decorate_result['structure']
            # Run seekpath on the decorated structure
            kpath_results = func(decorated, **inputs)
            decorated_primitive = kpath_results['primitive_structure']
            # Convert back to undecorated structures and add consistent magmom input
            dedecorate_result = magnetic_structure_dedecorate(decorated_primitive, decorate_result['mapping'])
            self.ctx.magmom = dedecorate_result['magmom'].get_list()
            self.ctx.current_structure = dedecorate_result['structure']
        else:
            kpath_results = func(self.ctx.current_structure, **inputs)
            self.ctx.current_structure = kpath_results['primitive_structure']

        if not np.allclose(self.ctx.current_structure.cell, current_structure_backup.cell):
            if self.inputs.scf.get('kpoints'):
                self.report(
                    'The primitive structure is not the same as the input structure but explicit kpoints are supplied'
                    ' - aborting the workchain.'
                )
                return self.exit_codes.ERROR_INPUT_STRUCTURE_NOT_PRIMITIVE  # pylint: disable=no-member
            self.report(
                'The primitive structure is not the same as the input structure - using the former for all calculations'
                ' from now.'
            )
        self.ctx.bs_kpoints = kpath_results['explicit_kpoints']
        self.out('primitive_structure', self.ctx.current_structure)
        if 'parameters' in kpath_results:
            self.out('seekpath_parameters', kpath_results['parameters'])

    def run_scf(self) -> None:
        """
        Run the SCF calculation
        """

        base_work = WorkflowFactory(self._base_wk_string)
        inputs = AttributeDict(self.exposed_inputs(base_work, namespace='scf'))
        inputs.metadata.call_link_label = 'scf'
        inputs.metadata.label = self.inputs.metadata.label + ' SCF'
        inputs.structure = self.ctx.current_structure

        # Turn off cleaning of the working directory
        if not inputs.get('keep_last_workdir', False):
            inputs.keep_last_workdir = orm.Bool(True)

        # Ensure that writing the CHGCAR file is on
        pdict = inputs.parameters.get_dict()
        if (pdict[OVERRIDE_NAMESPACE].get('lcharg') is False) or (pdict[OVERRIDE_NAMESPACE].get('LCHARG') is False):
            pdict[OVERRIDE_NAMESPACE]['lcharg'] = True
            inputs.parameters = orm.Dict(dict=pdict)
            self.report('Correction: setting LCHARG to True')

        # Take magmom from the context, in case that the magmom is rearranged in the primitive cell
        magmom = self.ctx.get('magmom')
        if magmom:
            inputs.parameters = update_nested_dict_node(inputs.parameters, {OVERRIDE_NAMESPACE: {'magmom': magmom}})

        running = self.submit(base_work, **inputs)
        self.report(f'Running SCF calculation {running}')
        self.to_context(workchain_scf=running)

    def verify_scf(self) -> Any:
        """Inspect the SCF calculation"""
        scf_workchain = self.ctx.workchain_scf
        if not scf_workchain.is_finished_ok:
            self.report('SCF workchain finished with Error')
            return self.exit_codes.ERROR_SUB_PROC_SCF_FAILED

        # Store the charge density or remote reference
        if 'chgcar' in scf_workchain.outputs:
            self.ctx.chgcar = scf_workchain.outputs.chgcar
        else:
            self.ctx.chgcar = None
        self.ctx.restart_folder = scf_workchain.outputs.remote_folder
        self.report(f'SCF calculation {scf_workchain} completed')

    def run_bands_dos(self) -> Any:
        """Run the bands and the DOS calculations"""
        base_work = WorkflowFactory(self._base_wk_string)

        # Use the SCF inputs as the base
        inputs = AttributeDict(self.exposed_inputs(base_work, namespace='scf'))
        inputs.structure = self.ctx.current_structure

        if self.ctx.restart_folder:
            inputs.restart_folder = self.ctx.restart_folder

        if self.ctx.chgcar:
            inputs.chgcar = self.ctx.chgcar

        if not (inputs.get('restart_folder') or inputs.get('chgcar')):
            raise RuntimeError('One of the restart_folder or chgcar must be set for non-scf calculations')

        running = {}

        only_dos = self.inputs.band_settings['only_dos']

        if only_dos is False:
            if 'bands' in self.inputs:
                bands_input = AttributeDict(self.exposed_inputs(base_work, namespace='bands'))
            else:
                bands_input = AttributeDict(
                    {
                        'settings': orm.Dict(dict={'parser_settings': {'include_node': ['bands']}}),
                        'parameters': orm.Dict(dict={'charge': {'constant_charge': True}}),
                    }
                )

            # Special treatment - combine the parameters
            parameters = inputs.parameters.get_dict()
            bands_parameters = bands_input.parameters.get_dict()

            if 'charge' in bands_parameters:
                bands_parameters['charge']['constant_charge'] = True
            else:
                bands_parameters['charge'] = {'constant_charge': True}

            update_nested_dict(parameters, bands_parameters)

            # Apply updated parameters
            inputs.update(bands_input)
            inputs.parameters = orm.Dict(dict=parameters)

            # Check if add_bands
            settings = inputs.get('settings')
            essential = {'parser_settings': {'include_node': ['bands']}}
            if settings is None:
                inputs.settings = orm.Dict(dict=essential)
            else:
                inputs.settings = update_nested_dict_node(settings, essential, extend_list=True)

            # Swap with the default kpoints generated
            inputs.kpoints = self.ctx.bs_kpoints

            # Tag the calculation
            inputs.metadata.label = self.inputs.metadata.label + ' BS'
            inputs.metadata.call_link_label = 'bs'

            bands_calc = self.submit(base_work, **inputs)
            running['bands_workchain'] = bands_calc
            self.report(f'Submitted workchain {bands_calc} for band structure')

        # Do DOS calculation if dos input namespace is populated or a
        # dos_kpoints input is passed.
        if (self.inputs.band_settings['run_dos']) or ('dos' in self.inputs):
            if 'dos' in self.inputs:
                dos_input = AttributeDict(self.exposed_inputs(base_work, namespace='dos'))
            else:
                dos_input = AttributeDict(
                    {
                        'parameters': orm.Dict(dict={'charge': {'constant_charge': True}}),
                    }
                )
                # Use the supplied kpoints density for DOS
                dos_kpoints = orm.KpointsData()
                dos_kpoints.set_cell_from_structure(self.ctx.current_structure)
                dos_kpoints.set_kpoints_mesh_from_density(self.inputs.band_settings['dos_kpoints_distance'] * 2 * np.pi)
                dos_input.kpoints = dos_kpoints

            # Special treatment - combine the parameters
            parameters = inputs.parameters.get_dict()
            dos_parameters = dos_input.parameters.get_dict()
            update_nested_dict(parameters, dos_parameters)

            # Ensure we start from constant charge
            if 'charge' in dos_parameters:
                dos_parameters['charge']['constant_charge'] = True
            else:
                dos_parameters['charge'] = {'constant_charge': True}

            # Apply updated parameters
            inputs.update(dos_input)
            inputs.parameters = orm.Dict(dict=parameters)

            if 'dos' not in self.inputs:
                # kindly add `add_dos` if the `dos` input namespace is not
                # explicitly defined.
                settings = inputs.get('settings')
                essential = {'parser_settings': {'include_node': ['dos', 'bands']}}

                if settings is None:
                    inputs.settings = orm.Dict(dict=essential)
                else:
                    inputs.settings = update_nested_dict_node(settings, essential, extend_list=True)

            # Set the label
            inputs.metadata.label = self.inputs.metadata.label + ' DOS'
            inputs.metadata.call_link_label = 'dos'

            dos_calc = self.submit(base_work, **inputs)
            running['dos_workchain'] = dos_calc
            self.report(f'Submitted workchain {dos_calc} for DOS')

        return self.to_context(**running)

    def inspect_bands_dos(self) -> Any:
        """Inspect the bands and dos calculations"""

        exit_code = None

        if 'bands_workchain' in self.ctx:
            bands = self.ctx.bands_workchain
            if not bands.is_finished_ok:
                self.report(f'Bands calculation finished with error, exit_status: {bands}')
                exit_code = self.exit_codes.ERROR_SUB_PROC_BANDS_FAILED
            self.out(
                'band_structure',
                compose_labelled_bands(bands.outputs.bands, bands.inputs.kpoints),
            )
        else:
            bands = None

        if 'dos_workchain' in self.ctx:
            dos = self.ctx.dos_workchain
            if not dos.is_finished_ok:
                self.report(f'DOS calculation finished with error, exit_status: {dos.exit_status}')
                exit_code = self.exit_codes.ERROR_SUB_PROC_DOS_FAILED

            # Attach outputs
            self.out('dos', dos.outputs.dos)
            if 'projectors' in dos.outputs:
                self.out('projectors', dos.outputs.projectors)
        else:
            dos = None

        return exit_code

    def on_terminated(self) -> None:
        """
        Clean the remote directories of all called childrens
        """

        super().on_terminated()

        if self.inputs.clean_children_workdir.value != 'none':
            cleaned_calcs = []
            for called_descendant in self.node.called_descendants:
                if isinstance(called_descendant, orm.CalcJobNode):
                    try:
                        called_descendant.outputs.remote_folder._clean()  # pylint: disable=protected-access
                        cleaned_calcs.append(called_descendant.pk)
                    except (OSError, KeyError):
                        pass

            if cleaned_calcs:
                self.report(f'cleaned remote folders of calculations: {" ".join(map(str, cleaned_calcs))}')


@calcfunction
def seekpath_structure_analysis(structure: orm.StructureData, band_settings: orm.Dict) -> Any:
    """Primitivize the structure with SeeKpath and generate the high symmetry k-point path through its Brillouin zone.
    This calcfunction will take a structure and pass it through SeeKpath to get the normalized primitive cell and the
    path of high symmetry k-points through its Brillouin zone. Note that the returned primitive cell may differ from the
    original structure in which case the k-points are only congruent with the primitive cell.
    The keyword arguments can be used to specify various Seekpath parameters, such as:

    - with_time_reversal: True
    - reference_distance: 0.025
    - recipe: 'hpkot'
    - threshold: 1e-07
    - symprec: 1e-05
    - angle_tolerance: -1.0

    Note that exact parameters that are available and their defaults will depend on your Seekpath version.
    """

    # All keyword arugments should be `Data` node instances of base type and so should have the `.value` attribute
    return get_explicit_kpoints_path(structure, **band_settings.get_dict())


@calcfunction
def compose_labelled_bands(bands: orm.BandsData, kpoints: orm.KpointsData) -> orm.BandsData:
    """
    Add additional information from the kpoints allow richer informations
    to be stored such as band structure labels.
    """
    new_bands = deepcopy(bands)
    new_bands.set_kpointsdata(kpoints)
    return new_bands


@calcfunction
def get_primitive_strucrture_and_scf_kpoints(structure: orm.StructureData) -> Any:
    """
    This function dryruns a VASP calculation using the primitive structure obtained by performing seekpath analyses

    The input StructureData should be returned by an VaspRelaxWorkChain which will be used for dryun using local
    VASP and getting the explicity kpoints for SCF calculation.
    """
    # Locate the relaxation work

    # Locate the relaxation work
    relax_work = structure.base.links.get_incoming(link_label_filter='relax__structure').one().node
    primitive = get_explicit_kpoints_path(structure)['primitive_structure']

    # Create an restart builder
    builder = relax_work.get_builder_restart()
    builder.structure = primitive

    # Dryrun and construct the SCF kpoints
    kpoint_weights = np.array(dryrun_relax_builder(builder)['kpoints_and_weights'])
    scf_kpoints = orm.KpointsData()
    scf_kpoints.set_kpoints(kpoint_weights[:, :3], weights=kpoint_weights[:, -1])
    return {'primitive': primitive, 'scf_kpoints': scf_kpoints}


class VaspHybridBandsWorkChain(VaspBandsWorkChain):
    """
    Bands workchain for hybrid calculations

    This workchain compute the bandstructure by adding band path segments as zero-weighted
    kpoints for self-consistent calculations. This is mainly for hybrid calculations, but can
    also be used for GGA calculations, although it would be not as efficient as the non-SCF
    method implemented in ``VaspBandsWorkChain``.

    In contrast to ``VaspBandsWorkChain`` this workflow requires and explicitly defined kpoints
    set for the ``scf.kpoints`` port. This can be obtained by parsing the ``IBZKPT`` file from
    and existing calculation or dryrun. Or by parsing the ``vasprun.xml`` file.

    If a relaxation workchain is run as part of the process, the ``kpoints`` output returned can
    be used for this purpose automatically.

    Only the `scf` namespace will be used for performing the calculation

    TODO:
     - Warn if the calculation is not actually a hybrid one
     - Automatic Kpoints from dryruns
    """

    @classmethod
    def define(cls, spec: Any) -> None:
        """Initialise the WorkChain class"""
        super().define(spec)
        relax_work = WorkflowFactory(cls._relax_wk_string)
        base_work = WorkflowFactory(cls._base_wk_string)

        spec.input('structure', help='The input structure', valid_type=orm.StructureData)
        spec.expose_inputs(
            relax_work,
            namespace='relax',
            exclude=('structure',),
            namespace_options={
                'required': False,
                'populate_defaults': False,
                'help': 'Inputs for Relaxation workchain, if needed',
            },
        )
        spec.expose_inputs(
            base_work,
            namespace='scf',
            exclude=('structure',),
            namespace_options={
                'required': True,
                'populate_defaults': True,
                'help': 'Inputs for SCF workchain, mandatory',
            },
        )
        spec.input(
            'clean_children_workdir',
            valid_type=orm.Str,
            serializer=to_aiida_type,
            help='What part of the called children to clean',
            required=False,
            default=lambda: orm.Str('none'),
        )
        spec.outline(
            cls.setup,
            if_(cls.should_do_relax)(
                cls.run_relax,
                cls.verify_relax,
            ),
            if_(cls.should_generate_path)(cls.generate_path),
            # Find the SCF kpoints from geometry optimisation
            if_(cls.no_scf_kpoints)(cls.get_scf_kpoints_relax),
            # Generate the SCF kpoints using spglib
            if_(cls.no_scf_kpoints)(cls.get_scf_kpoints_spglib),
            # If the above fails, we need to run a SCF calculation to get the kpoints
            if_(cls.no_scf_kpoints)(cls.run_scf_for_kpoints, cls.verify_scf_for_kpoints),
            cls.make_splitted_kpoints,  # Split the kpoints
            cls.run_scf_multi,  # Launch split calculation
            cls.inspect_and_combine_bands,  # Combined the band structure
        )

        spec.output(
            'primitive_structure',
            required=False,
            help='Primitive structure used for band structure calculations',
        )
        spec.output('band_structure', required=False, help='Computed band structure with labels')
        spec.output('seekpath_parameters', help='Parameters used by seekpath', required=False)

        spec.exit_code(501, 'ERROR_SUB_PROC_RELAX_FAILED', message='Relaxation workchain failed')
        spec.exit_code(502, 'ERROR_SUB_PROC_SCF_FAILED', message='SCF workchain failed')
        spec.exit_code(
            503,
            'ERROR_SUB_PROC_BANDS_FAILED',
            message='Band structure workchain failed',
        )
        spec.exit_code(504, 'ERROR_SUB_PROC_DOS_FAILED', message='DOS workchain failed')
        spec.exit_code(
            505,
            'ERROR_NO_VALID_SCF_KPOINTS_INPUT',
            message='Cannot found valid inputs for SCF kpoints',
        )
        spec.exit_code(
            601,
            'ERROR_INPUT_STRUCTURE_NOT_PRIMITIVE',
            message='The input structure is not the primitive one!',
        )

    def setup(self) -> None:
        super().setup()
        self.ctx.scf_kpoints = None
        if 'kpoints' in self.inputs.scf:
            self.ctx.scf_kpoints = self.inputs.scf.kpoints

    def no_scf_kpoints(self) -> bool:
        """Check if the kpoints for SCF has NOT been set"""
        if self.ctx.scf_kpoints is None:
            return True
        return False

    def get_scf_kpoints_relax(self) -> None:
        """Try extract SCF kpoints from relaxation workchain"""

        if 'workchain_relax' not in self.ctx:
            self.report('No workchain_relax found in context - skip extracting scf kpoints from previous calculation')
            return

        # Check if symmetry is consistent
        incar_relax = self.ctx['workchain_relax'].inputs.vasp.parameters['incar']
        incar_scf = self.inputs.scf.parameters['incar']
        if incar_scf.get('isym', 2) != incar_relax.get('isym', 2):
            self.report(
                'The symmetry of the SCF calculation is not consistent with the relaxation calculation. '
                'Cannot reuse the IBZ kpoitns from the relaxation calculation.'
            )
            return

        # Check if the scf_kpoints is consistent with the current structure
        relaxed = self.ctx.workchain_relax.outputs.relax.structure
        kpt_cell = np.array(relaxed.cell)
        current_cell = np.array(self.ctx.current_structure.cell)
        assert kpt_cell.shape == (3, 3)
        assert current_cell.shape == (3, 3)
        # Compute the cell parameters
        par1 = cell_to_cellpar(kpt_cell)
        par2 = cell_to_cellpar(current_cell)
        if not np.allclose(par1, par2, 1e-5):
            self.report(
                'Cell of the last relaxation step does not match the current structure.SCF kpoints cannot be used'
            )
            return

        # Extract the kpoints from the relaxation calculation.
        if 'kpoints' in self.ctx['workchain_relax'].outputs:
            self.report(f'Using output from <{self.ctx.workchain_relax}> for SCF kpoints.')
            self.ctx.scf_kpoints = self.ctx.workchain_relax.outputs.kpoints
        else:
            # Parse from relaxation output
            # Try getting the kpoints from the retrieved folder
            self.report(f'Extracted SCF kpoints from retrieved vasprun.xml of <{self.ctx.workchain_relax}>.')
            self.ctx.scf_kpoints = extract_kpoints_from_calc(self.ctx.workchain_relax)

    def get_scf_kpoints_spglib(self) -> None:
        """
        Generate SCF kpoints using spglib
        """
        incar_scf = self.inputs.scf.parameters['incar']
        magmom = incar_scf.get('magmom', None)
        symmetry_reduce = incar_scf.get('isym', 2) > 0
        if symmetry_reduce and magmom is not None:
            # Check if symmetry is broken by magnetic moments
            species = self.inputs.structure.get_ase().get_chemical_symbols()
            if isinstance(magmom, str):
                magmom = magmom.split()
            assert len(magmom) == len(species), (
                f'Mismatch between the magmom ({len(magmom)}) and the number of atoms ({len(species)}).'
            )
            if len(set(zip(magmom, species))) != len(set(species)):
                self.report('Symmetry is broken by magnetic magmoms, cannot use spglib to generate the kpoints')
                return
        kpt = get_ir_kpoints_data(
            self.ctx.current_structure,
            self.inputs.scf.kpoints_spacing * np.pi * 2,  # Note that aiida-vasp assumes a 2pi factor in the unit
            symprec=self.inputs.band_settings.get('symprec', 1e-5),
            symmetry_reduce=symmetry_reduce,
        )
        self.ctx.scf_kpoints = kpt

    def make_splitted_kpoints(self) -> Any:
        """Split the kpoints"""
        # Fully specified band structure kpoints
        full_kpoints = self.ctx.bs_kpoints

        scf_kpoints = self.ctx.scf_kpoints
        if scf_kpoints is None:
            self.report('No valid SCF kpoints is avaliable to use. Please define scf.kpoints explicitly!')
            return self.exit_codes.ERROR_NO_VALID_SCF_KPOINTS_INPUT  # pylint: disable=no-member

        # Number of kpoints per split, NOT including the SCF kpoints
        nscf = scf_kpoints.get_kpoints().shape[0]
        per_split = orm.Int(self.inputs.band_settings['kpoints_per_split'] - nscf)
        if (per_split / nscf) <= 0.5:
            per_split = int(nscf * 0.5)
            self.report(f'WARNING: Too few actual band k points per split, setting it to: {per_split + nscf}')
        kpoints_for_calc = split_kpoints(scf_kpoints, full_kpoints, per_split)
        self.ctx.kpoints_for_calc = kpoints_for_calc

    def should_do_scf_for_scf_kpoints(self) -> bool:
        """Check if one should redo a SCF run to obtain the IBZKPT"""
        scf_kpoints = self.get_scf_kpoints()
        if scf_kpoints is None:
            return True

        # Check if the scf_kpoints is consistent with the current structure
        scf_cell = np.array(scf_kpoints.cell)
        current_cell = np.array(self.ctx.current_structure.cell)
        assert scf_cell.shape == (3, 3)
        assert current_cell.shape == (3, 3)
        # Compute the cell parameters
        par1 = cell_to_cellpar(scf_cell)
        par2 = cell_to_cellpar(current_cell)

        # If the cells parameters are different - we should run a SCF calculation to get the kpoint for this new cell
        if not np.allclose(par1, par2, 1e-5):
            return True
        return False

    def run_scf_for_kpoints(self) -> Any:
        """
        Run an SCF calculation to just obtain the kpoints for the current structure
        Ideally we should do this in a dryrun mode @ local machine
        """
        workflow_class = WorkflowFactory(self._base_wk_string)

        # Check if we need to turn off spin polarization
        inputs = self.exposed_inputs(workflow_class, 'scf')
        pdict = inputs.parameters.get_dict()

        # Reuse the wavecar if requested
        if self.inputs.band_settings.get('hybrid_reuse_wavecar', False):
            self.report('Setting ISTART=1 to reuse WAVECAR from the previous calculation.')
            pdict['incar']['istart'] = 1
            inputs.parameters = orm.Dict(pdict)

        # Ensure that the kpoints are returned by the parser
        if 'settings' not in inputs:
            inputs.settings = orm.Dict(dict={'parser_settings': {'include_node': ['kpoints']}})
        else:
            # Merge with 'parser_settings'
            inputs.settings = update_nested_dict_node(
                inputs.settings, {'parser_settings': {'include_node': ['kpoints']}}, extend_list=True
            )

        inputs.metadata.label = self.inputs.metadata.label + ' SCF KPOINTS'
        inputs.metadata.call_link_label = 'scf_for_kpoints'
        inputs.structure = self.ctx.current_structure  # Use the current structure as reference
        running = self.submit(workflow_class, **inputs)
        return self.to_context(workchain_scf_for_kpoints=running)

    def verify_scf_for_kpoints(self) -> Any:
        """Inspect the SCF for kpoints calculation"""
        scf_workchain = self.ctx.workchain_scf_for_kpoints
        if not scf_workchain.is_finished_ok:
            self.report('SCF for kpoints workchain finished with Error')
            return self.exit_codes.ERROR_SUB_PROC_SCF_FAILED

        # Save the obtained kpoints
        self.ctx.scf_kpoints = scf_workchain.outputs.kpoints
        self.report(f'SCF calculation {scf_workchain} completed and obtained kpoints {self.ctx.scf_kpoints}')

    def run_scf_multi(self) -> Any:
        """
        Launch multiple SCF calculations with zero-weighted kpoints for segments of the band structure
        """

        workflow_class = WorkflowFactory(self._base_wk_string)

        # Check if we need to turn off spin polarization
        inputs = self.exposed_inputs(workflow_class, 'scf')
        pdict = inputs.parameters.get_dict()
        # Check if we really need to run spin polarized calculation
        relax_work = self.ctx.get('workchain_relax', None)
        if relax_work is not None and pdict.get('incar', {}).get('ispin') == 2:
            self.report('Checking the magnetization of the relaxed structure.')
            # Check if the site magnetizations are all zero
            mag = relax_work.outputs.misc.get('site_magnetization')
            if not _is_magnetic_via_site_moment(mag):
                pdict['incar']['ispin'] = 1
                self.report('Turnning off spin polarization for band structure calculation for non-magnetic system.')
                inputs.parameters = orm.Dict(pdict)

        # Reuse the wavecar if requested
        if self.inputs.band_settings.get('hybrid_reuse_wavecar', False):
            self.report('Setting ISTART=1 to reuse WAVECAR from the previous calculation.')
            pdict['incar']['istart'] = 1
            inputs.parameters = orm.Dict(pdict)

        pnode = inputs.parameters

        for key, value in self.ctx.kpoints_for_calc.items():
            idx = int(key.split('_')[-1])

            inputs = self.exposed_inputs(workflow_class, 'scf')
            # Use the updated parameters
            inputs.parameters = pnode

            if self.inputs.band_settings.get('hybrid_reuse_wavecar', False):
                inputs.restart_folder = relax_work.outputs.remote_folder

            # Ensure that the bands are parsed
            if 'settings' not in inputs:
                inputs.settings = orm.Dict(dict={'parser_settings': {'include_node': ['bands']}})
            else:
                # Merge with 'parser_settings'
                inputs.settings = update_nested_dict_node(
                    inputs.settings, {'parser_settings': {'include_node': ['bands']}}, extend_list=True
                )

            # Swap the kpoints the the one with zero-weight parts
            inputs.kpoints = value
            inputs.metadata.label = self.inputs.metadata.label + f' SPLIT {idx}'
            inputs.metadata.call_link_label = f'bandstructure_split_{idx:03d}'
            inputs.structure = self.ctx.current_structure
            running = self.submit(workflow_class, **inputs)
            self.report(f'launching {workflow_class.__name__}<{running.pk}> for split #{idx}')
            self.to_context(workchains=append_(running))

    def inspect_and_combine_bands(self) -> None:
        """
        Inspect that all calculations have finished OK
        """
        workchains = self.ctx.workchains

        return_codes = [work.exit_status for work in workchains]
        if any(return_codes):
            self.report('At least one calculation did not have zero return code!')

        # Extract the bands information
        self.report(f'Extracting output bandstructure from {len(self.ctx.workchains)} workchains.')
        kwargs = {}
        for work in workchains:
            link_label = work.base.links.get_incoming(link_type=LinkType.CALL_WORK).one().link_label
            link_idx = int(link_label.split('_')[-1])
            kwargs[f'band_{link_idx:03d}'] = work.outputs.bands
            kwargs[f'kpoint_{link_idx:03d}'] = work.inputs.kpoints

        combined_bands = combine_bands_data(self.ctx.bs_kpoints, **kwargs)
        self.out('band_structure', combined_bands)


@calcfunction
def split_kpoints(scf_kpoints: orm.KpointsData, band_kpoints: orm.KpointsData, kpn_per_split: orm.Int) -> Any:
    """
    Split the kpoints into multiple one and combined with SCF kpoints

    The kpoints for band structure calculation has zero weights
    """
    return _split_kpoints(scf_kpoints, band_kpoints, kpn_per_split)


def _split_kpoints(scf_kpoints: orm.KpointsData, band_kpoints: orm.KpointsData, kpn_per_split: orm.Int) -> Any:
    """
    Split the kpoints into multiple one and combined with SCF kpoints

    The kpoints for band structure calculation has zero weights
    """
    scf_kpoints_array, scf_weights_array = scf_kpoints.get_kpoints(also_weights=True)
    band_kpn = band_kpoints.get_kpoints()
    nband_kpts = band_kpn.shape[0]
    nscf_kpts = scf_kpoints_array.shape[0]

    # Split the kpoints
    kpn_per_split = int(kpn_per_split)
    kpt_splits = [band_kpn[i : i + kpn_per_split] for i in range(0, nband_kpts, kpn_per_split)]

    splitted_kpoints = {}
    for isplit, skpts in enumerate(kpt_splits):
        kpt = orm.KpointsData()
        kpt_array = np.concatenate([scf_kpoints_array, skpts], axis=0)
        weights_array = np.zeros(kpt_array.shape[0])
        # Set the weights for SCF kpoints
        weights_array[:nscf_kpts] = scf_weights_array
        # Set kpoints and the weights
        kpt.set_kpoints(kpt_array, weights=weights_array)
        kpt.label = f'SPLIT {isplit:03d}'
        kpt.description = 'Splitted kpoints'
        splitted_kpoints[f'bs_kpoints_{isplit:03d}'] = kpt

    return splitted_kpoints


def dryrun_split_kpoints(
    structure: orm.StructureData,
    scf_kpoints: orm.KpointsData,
    kpn_per_split: orm.Int,
    kpoints_args: Any = None,
    verbose: bool = True,
) -> Any:
    """
    Perform a "dryrun" for splitting the kpoints
    """
    if kpoints_args is None:
        kpoints_args = {}
    seekpath_results = get_explicit_kpoints_path(structure, **kpoints_args)
    explicit_kpoints = seekpath_results['explicit_kpoints']
    splitted = _split_kpoints(scf_kpoints, explicit_kpoints, kpn_per_split)
    if verbose:
        nseg = len(splitted)
        nkpts = [kpn.get_kpoints().shape[0] for kpn in splitted.values()]
        print(f'Splitted in to {nseg} segements with number of kpoints: {nkpts}')
    return seekpath_results, splitted


@calcfunction
def combine_bands_data(bs_kpoints: orm.KpointsData, **kwargs: Any) -> orm.BandsData:
    """
    Combine splitted bands and kpoints data

    The inputs should be supplied as keyword arguments like `band_001`, `kpoint_001` for the splitted
    kpoints and correspdonging bands data from each calculation.
    The `bs_kpoints` is the originally generated band structure path.

    Returns a `BandsData` by combining the zero-weighted bands from each calculation.
    """
    kpoints_list = [[node, int(key.split('_')[1])] for key, node in kwargs.items() if 'kpoint' in key]
    kpoints_list.sort(key=lambda x: x[1])
    kpoints_list = [item[0] for item in kpoints_list]

    bands_list = [[node, int(key.split('_')[1])] for key, node in kwargs.items() if 'band' in key]
    bands_list.sort(key=lambda x: x[1])
    bands_list = [item[0] for item in bands_list]

    return _combine_bands_data(bs_kpoints, kpoints_list, bands_list)


def _combine_bands_data(
    bs_kpoints: orm.KpointsData,
    kpoints_list: List[orm.KpointsData],
    bands_list: List[orm.BandsData],
) -> orm.BandsData:
    """
    Combine bands from splitted kpoints into a single bands node.

    The list of kpoints and bands must be sorted in the right order.
    """
    bands_array_combine = []
    occu_array_combine = []
    kpoints_combine = []
    fermi_levels = []

    for skpts, sbands in zip(kpoints_list, bands_list):
        fermi_levels.append(sbands.base.attributes.get('fermi_level', None))
        kpt_array, weights_array = skpts.get_kpoints(also_weights=True)
        zero_weight_mask = weights_array == 0.0
        kpoints_combine.append(kpt_array[zero_weight_mask, :])

        bands_array = sbands.get_bands()
        if 'occupations' in sbands.get_arraynames():
            occ_array = sbands.get_array('occupations')
        else:
            occ_array = None

        # Bands array can have three or two dimensions, we have to handle it separately
        if bands_array.ndim == 3:
            bands_array_combine.append(bands_array[:, zero_weight_mask, :])
            if occ_array is not None:
                occu_array_combine.append(occ_array[:, zero_weight_mask, :])
        else:
            bands_array_combine.append(bands_array[zero_weight_mask, :])
            if occ_array is not None:
                occu_array_combine.append(occ_array[zero_weight_mask, :])

    # Concatenate arrays
    if bands_array.ndim == 3:
        band_array_full = np.concatenate(bands_array_combine, axis=1)
        if occu_array_combine:
            occu_array_full = np.concatenate(occu_array_combine, axis=1)
        else:
            occu_array_full = None
    else:
        band_array_full = np.concatenate(bands_array_combine, axis=0)
        if occu_array_combine:
            occu_array_full = np.concatenate(occu_array_combine, axis=0)
        else:
            occu_array_full = None

    # Sanity check all valid kpoints should combine into the original path
    all_kpoints = np.concatenate(kpoints_combine, axis=0)
    if not np.allclose(all_kpoints, bs_kpoints.get_kpoints()):
        raise ValueError('The k-path segements do not much the original path when combined!')

    # Compose the node
    band_data = orm.BandsData()
    band_data.set_kpointsdata(bs_kpoints)
    band_data.set_bands(band_array_full, occupations=occu_array_full)
    # Set the fermi level of the combined bands
    if any(x is None for x in fermi_levels) or any(abs(entry - fermi_levels[0]) > 0.01 for entry in fermi_levels):
        logger.warning(
            f'Fermi level of the splitted calculations ({fermi_levels}) are not consistent! '
            'Using the first one as the combined fermi level.'
        )
    band_data.base.attributes.set('fermi_level', fermi_levels[0])
    band_data.base.attributes.set('efermi', fermi_levels[0])  # Alias for fermi level

    return band_data


def extract_kpoints_from_calc(calc: Any) -> Any:
    """
    Extract computed kpoints from a existing calculation
    """
    retrieved = calc.outputs.retrieved
    return extract_kpoints_from_retrieved(retrieved)


@calcfunction
def extract_kpoints_from_retrieved(retrieved: orm.FolderData) -> orm.KpointsData:
    """
    Extract explicity kpoints from a finished calculation
    """
    return _extract_kpoints_from_retrieved(retrieved)


def _extract_kpoints_from_retrieved(retrieved: orm.FolderData) -> orm.KpointsData:
    """
    Extract explicity kpoints from a finished calculation
    """

    with retrieved.base.repository.open('vasprun.xml', 'rb') as fh:
        parser = VasprunParser(handler=fh)

    vkpoints = parser.kpoints
    if vkpoints['mode'] != 'explicit':
        raise ValueError('Only explicity kpoints is supported!')

    kpoints_array = vkpoints['points']
    weights_array = vkpoints['weights']

    kpoints_data = orm.KpointsData()
    kpoints_data.set_kpoints(kpoints=kpoints_array, weights=weights_array)
    kpoints_data.set_cell(parser.structure['unitcell'])

    return kpoints_data


def _is_magnetic_via_site_moment(mag: Any) -> bool:
    has_mag = False
    # Iterate over dictionaries of the site moments of each site
    for site in mag['sphere']['x']['site_moment'].values():
        # Check if any of the moments is non-zero
        if any(abs(x) > SITE_MAG_THRESHOLD for x in site.values()):
            has_mag = True
            break
    return has_mag



"""Test the BaseFileParser."""
# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import,protected-access

import pytest
from aiida import orm

from aiida_vasp.parsers.content_parsers.base import BaseFileParser


def test_base_content_parser(tmpdir):
    """Test functionality of the BaseFileParser, that is not already covered by any of the subclass tests."""
    with pytest.raises(TypeError):
        # Need to be initialized with either the `handler` or `data` parameters
        _ = BaseFileParser()
    temp_path = tmpdir / 'dummyfile'
    with pytest.raises(NotImplementedError):
        # Does not have a init_handler method, should be implemented in child
        with open(temp_path, 'w', encoding='utf8') as handler:
            _ = BaseFileParser(handler=handler)
    with pytest.raises(NotImplementedError):
        # Does not have a init_data method, should be implemented in child
        _ = BaseFileParser(data=orm.Dict(dict={}))
    with pytest.raises(TypeError):
        # Check that something else than Data for the AiiDA data type raises an exception.
        _ = BaseFileParser(data=[])



"""
Representation of WAVECAR and WAVEDER objects.

Wave function data node (stores WAVECAR and WAVEDER objects in the repository).
"""

# pylint: disable=abstract-method, too-few-public-methods, no-name-in-module
# explanation: pylint wrongly complains about (aiida) Node not implementing query
from aiida.orm import SinglefileData


class WavefunData(SinglefileData):
    pass



"""Test the EIGENVAL parser."""
# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import

import numpy as np
import pytest


@pytest.mark.parametrize(['eigenval_parser'], [('eigenval',)], indirect=True)
def test_parse_eigenval(eigenval_parser):
    """Parse a reference EIGENVAL with the EigenvalParser and compare the result to a reference."""
    result = eigenval_parser.get_quantity('eigenval-eigenvalues')
    eigenvalues = np.array(
        [
            [
                [
                    -1.439825,
                    2.964373,
                    2.964373,
                    2.964373,
                    7.254542,
                    7.254542,
                    7.254542,
                    11.451811,
                    11.670398,
                    11.670398,
                ]
            ]
        ]
    )
    metadata = {
        0: [4, 4, 1, 1],
        1: [16.48482, 4.04e-10, 4.04e-10, 4.04e-10, 1e-16],
        2: 0.0001,
        'n_ions': 4,
        'n_atoms': 4,
        'p00': 1,
        'nspin': 1,
        'cartesian': True,
        'name': 'unknown system',
        'some_num': 12,
        'n_bands': 10,
        'n_kp': 1,
    }
    assert result['metadata'] == metadata
    assert np.allclose(result['eigenvalues'], eigenvalues)



"""
Base classes for the VASP content parsers.

===============================

Contains the base classes for the VASP content parsers.
"""
# pylint: disable=import-outside-toplevel

from __future__ import annotations

from typing import Any, BinaryIO, TextIO

from aiida.common import AIIDA_LOGGER
from aiida.orm import Data


class BaseFileParser:
    """
    Base class for all the content parsers which parse (read and write) VASP files.

    This class provides the interface for parsing and writing VASP files using parsevasp.
    It ensures integration with the AiiDA parsing framework and preparation before submission.
    Specific content parser interfaces should inherit from this class.

    There are two main usage patterns:

    1. **Parsing from file**: Initialize with ``handler`` (a file-like object), then fetch quantities
       using :meth:`get_quantity` with the appropriate key. The valid keys are defined in the
       ``PARSABLE_QUANTITIES`` class parameter of each subclass.

    2. **Writing from data**: Initialize with ``data`` (an AiiDA data node), then use :meth:`get_quantity`
       to retrieve the same node, or use :meth:`write` to write the content to a file.

    :param handler: File-like object containing content to be parsed. Used when parsing completed calculations.
    :type handler: file-like object, optional
    :param data: AiiDA data node. Used when writing VASP input files.
    :type data: object, optional
    :param settings: Parser settings, e.g. which quantities to compose into nodes.
    :type settings: dict, optional
    :param options: Parser options, e.g. extra options for content parsers such as selective dynamics for POSCAR.
    :type options: dict, optional
    """

    OPEN_MODE: str = 'r'
    PARSABLE_QUANTITIES: dict[str, Any] = {}
    DEFAULT_SETTINGS: dict[str, Any] = {'quantities_to_parse': []}

    def __init__(
        self,
        *,
        handler: TextIO | BinaryIO | None = None,
        data: Data | None = None,
        settings: dict[str, Any] | None = None,
        options: dict[str, Any] | None = None,
    ) -> None:
        super().__init__()
        # Make sure we only accept initialization with either ``handler`` or ``data``.
        if (handler is not None and data is not None) or (handler is None and data is None):
            raise TypeError('Supply at bare minimum either argument handler or data to initialize parser.')

        # Make sure logger messages in the parser are passed to the AiiDA logger.
        self._logger = AIIDA_LOGGER.getChild(self.__class__.__name__)

        # What quantities the specific content parser can provide.
        self._parsable_quantities = self.PARSABLE_QUANTITIES
        # The container for the parsed data when the ``get_quantity`` is executed, i.e. in the node composer
        # at a later stage.
        self._parsed_content = {}
        # The content parser, which will be an instance of one of the parsevasp parser classes.
        self._content_parser = None
        # Content data, which is an AiiDA data structure.
        self._content_data = None
        # Parser settings.
        self._set_settings(settings)
        # Parser options.
        self._options = options

        self.parser_notifications = {}

        # Set ``handler`` (parsing from some source) or ``data`` (eventually for example executing write)
        if handler is not None:
            self._init_from_handler(handler)
        if data is not None:
            # Check that the supplied handler is Data, one of the AiiDA supported data types.
            if isinstance(data, Data):
                self._init_from_data(data)
            else:
                raise TypeError('The supplied handler is not of Data type.')

    def get_all_quantities(self) -> tuple[dict[str, Any], dict[str, Any]]:
        """
        Fetch all quantities that can be parsed.

        :return: Tuple of (parsed, errored) dictionaries.
        :rtype: tuple
        """
        parsed = {}
        errored = {}
        for name in self.PARSABLE_QUANTITIES:
            try:
                parsed[name] = getattr(self, name)
            except (TypeError, ValueError, KeyError, AttributeError, IndexError) as error:
                errored[name] = error
                continue
        return parsed, errored

    @property
    def parsable_quantities(self) -> list[str]:
        """
        Fetch the quantities that this content parser can provide.

        :return: List of parsable quantity keys.
        :rtype: list
        """
        return self._parsable_quantities

    def _set_settings(self, settings: dict[str, Any] | None) -> None:
        """
        Set the settings to be used for the content parser.

        :param settings: The settings to be used for the content parser. Can be None if no settings
          is supplied to init. Defaults are then set.
        :type settings: None or dict
        """
        if settings is None:
            # Apply defaults
            self._settings = self.DEFAULT_SETTINGS
        else:
            self._settings = settings
        if not self._settings.get('quantities_to_parse'):
            # Did not find any quantities to parse in the settings, set it to the default
            # for each content parser or to an empty list of not defined
            self._settings['quantities_to_parse'] = (
                self.DEFAULT_SETTINGS['quantities_to_parse'] if self.DEFAULT_SETTINGS.get('quantities_to_parse') else []
            )
        # Let us make sure the quantities to parser is in a list form
        if not isinstance(self._settings.get('quantities_to_parse'), list):
            raise TypeError('The quantities_to_parse is not defined as a list of quantities.')

    def get_quantity(self, quantity_key: str) -> Any:
        """
        Fetch the required quantity from the content parser.

        Either fetch it from an existing AiiDA data structure, a parsed content dictionary
        if that exists, otherwise parse this specific quantity using the loaded instance,
        which is now a specific content parser.

        :param quantity_key: Key of the quantity to be fetched.
        :type quantity_key: str
        :return: The requested quantity, or None if not parsable.
        :rtype: object or None
        """
        if self._content_data is not None:
            # If we have already set an AiiDA data structure, return it.
            # This is straightforward in our case as there is for the PoscarParser,
            # KpointsParser a 1:1 mapping between the parser and the AiiDA data (if
            # we ignore conversions between representations etc.).
            return self._content_data
        # We continue assuming we need to parse this quantity
        if quantity_key not in self._parsable_quantities:
            # Check if this quantity can be parsed by this content parser, if not,
            # return None.
            return None
        if self._parsed_content.get(quantity_key) is None:
            # Parsed content does not contain this quantity,
            # most likely none of the content is parsed. Parse
            # all relevant content now and store.
            self._parsed_content = self._parse_content()

        return self._parsed_content.get(quantity_key)

    def write(self, path: str) -> None:
        """
        Write VASP content to file using the loaded content parser.

        :param path: Path to write the file to.
        :type path: str
        """

        if self._content_parser is None or self._content_data is None:
            # Only write if we have an AiiDA data structure or parser prepared.
            if self._content_parser is None:
                # If we do not have a parser loaded before write, we have an
                # AiiDA data structure. Make sure that is on the form parsevasp expects
                # by initializing the content parser instance.
                self._content_parser = self._content_data_to_content_parser()
            # Now a content parser should be ready and its content can be
            # written using parsevasp. But the content parser could still be None
            # if there is something with the data that can not be parsed.
            with open(path, 'w', encoding='utf8') as handler:
                self._content_parser.write(file_handler=handler)
        else:
            raise ValueError('The content parser has not been initialized or no AiiDA data structure is preparred.')

    def _init_from_handler(self, handler: TextIO | BinaryIO) -> None:
        """
        Initialize using a file-like object.

        Should be overridden in specific content parsers under ``content_parsers``
        if it will accept parsable content.

        :param handler: File-like object that provides the necessary content to be parsed.
        :type handler: object
        """

        raise NotImplementedError(f'{self.__class__.__name__} does not implement a _init_from_handler() method.')

    def _init_from_data(self, data: Data) -> None:
        """
        Initialize using an AiiDA data structure.

        Should be overridden in specific content parsers under ``content_parsers``
        if it will accept an AiiDA data structure. It should also check that the
        right structure is supplied.

        :param data: A valid AiiDA data structure object.
        :type data: object
        """

        raise NotImplementedError(f'{self.__class__.__name__} does not implement a _init_from_data() method.')

    def _content_data_to_content_parser(self) -> Any:
        """
        Convert an AiiDA data structure to a content parser instance relevant for that
        data structure. E.g. ``Poscar`` from ``parsevasp`` for an AiiDA ``StructureData``.

        Should be overridden in specific content parsers under ``content_parsers``
        if it will accept an AiiDA data structure. It should also check that the
        right structure is supplied.

        :return: Instance of a content parser from ``parsevasp``, e.g. ``Poscar``.
        :rtype: object
        """

        raise NotImplementedError(
            f'{self.__class__.__name__} does not implement a _content_data_to_content_parser() method.'
        )

    def _parse_content(self) -> dict[str, Any]:
        """
        Parse the quantities configured and parseable from the content.

        :return: Dictionary of parsed quantities.
        :rtype: dict
        """
        quantities_to_parse = self._settings.get('quantities_to_parse')
        result = {}
        if self._content_parser is None:
            # Parsevasp threw an exception, which means content could not be parsed.
            for quantity in quantities_to_parse:
                if quantity in self._parsable_quantities:
                    result[quantity] = None
            return result

        for quantity in quantities_to_parse:
            if quantity in self._parsable_quantities:
                # In case there is a - in the quantity, we assume we can
                # parse this quantity from multiple sources, remove source as we do not want to used
                # the source in the property name, i.e. use last element in the split
                quantity_splitted = quantity.split('-')
                quantity_splitted = quantity_splitted[-1]
                result[quantity] = getattr(self, quantity_splitted)

        return result



"""
Immigrant calculation.


Enables the immigration of  externally run VASP calculations into AiiDA.

NOTE: This module is no longer maintained and not working.
"""

# pylint: disable=abstract-method, import-outside-toplevel, cyclic-import
# explanation: pylint wrongly complains about (aiida) Node not implementing query
from pathlib import Path
from typing import Any

import plumpy
from aiida import orm
from aiida.common import InputValidationError
from aiida.common.extendeddicts import AttributeDict
from aiida.common.folders import SandboxFolder
from aiida.common.lang import override
from aiida.common.links import LinkType
from aiida.engine import ProcessBuilder, ProcessSpec
from aiida.engine.processes.calcjobs.tasks import RETRIEVE_COMMAND

from aiida_vasp.calcs.vasp import VaspCalculation
from aiida_vasp.data.chargedensity import ChargedensityData
from aiida_vasp.data.potcar import PotcarData
from aiida_vasp.data.wavefun import WavefunData
from aiida_vasp.parsers.content_parsers.incar import IncarParser
from aiida_vasp.parsers.content_parsers.kpoints import KpointsParser
from aiida_vasp.parsers.content_parsers.poscar import PoscarParser
from aiida_vasp.parsers.content_parsers.potcar import MultiPotcarIo
from aiida_vasp.utils.aiida_utils import cmp_get_transport

# _IMMIGRANT_EXTRA_KWARGS = """
# vasp.vasp specific kwargs:

# :param use_chgcar: bool, if True, use the CHGCAR (has to exist) and convert it to an input node.
# :param use_wavecar: bool, if True, use the WAVECAR (has to exist) and convert it to an input node.
# """

# @update_docstring('immigrant', _IMMIGRANT_EXTRA_KWARGS, append=True)


class VaspImmigrant(VaspCalculation):
    """
    Parse VASP output objects stored in a specified directory.

    Simulate running the VaspCalculation up to the point where it can be
    retrieved and parsed, then hand over control to the runner for the rest.

    Immigrant calculation can be perfomed as follows::

       code = Code.get_from_string('vasp@local')
       folder = '/home/username/vasp-calc-dir'
       settings = {'parser_settings': {'add_energies': True,
                                       'add_forces': True,
                                       'electronic_step_energies': True}}
       VaspImmigrant = CalculationFactory('vasp.immigrant')
       builder = VaspImmigrant.get_builder_from_folder(code,
                                                       folder,
                                                       settings=settings)
       submit(builder)

    Instead of ``builder``, inputs dict is obtained similarly as::

       code = Code.get_from_string('vasp@local')
       folder = '/home/username/vasp-calc-dir'
       settings = {'parser_settings': {'add_energies': True,
                                       'add_forces': True,
                                       'electronic_step_energies': True}}
       VaspImmigrant = CalculationFactory('vasp.immigrant')
       inputs = VaspImmigrant.get_inputs_from_folder(code,
                                                     folder,
                                                     settings=settings)
       submit(VaspImmigrant, **inputs)

    The default metadata is set automatically as follows::

       {'options': {'max_wallclock_seconds': 1,
        'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 1}}}

    Specific scheduler may require setting ``resources`` differently
    (e.g., sge ``'parallel_env'``).

    ``get_inputs_from_folder`` and ``get_builder_from_folder`` accept several
    kwargs, see the docstring of ``get_inputs_from_folder``.

    """

    @classmethod
    def define(cls, spec: ProcessSpec) -> None:
        super().define(spec)
        spec.input('remote_workdir', valid_type=str, required=False, non_db=True)

    @override
    def run(self) -> plumpy.Wait:
        _ = super().run()

        # Make sure the retrieve list is set (done in presubmit so we need to call that also)
        with SandboxFolder() as folder:
            self.presubmit(folder)

        if 'remote_workdir' not in self.inputs:
            raise InputValidationError('immigrant calculations need inputs.remote_workdir.')

        self.node.set_remote_workdir(self.inputs.remote_workdir)  # pylint: disable=protected-access
        remotedata = orm.RemoteData(computer=self.node.computer, remote_path=self.inputs.remote_workdir)
        remotedata.base.links.add_incoming(self.node, link_type=LinkType.CREATE, link_label='remote_folder')
        remotedata.store()

        return plumpy.Wait(msg='Waiting to retrieve', data=RETRIEVE_COMMAND)

    @classmethod
    def get_inputs_from_folder(cls, code: orm.Code, remote_workdir: str, **kwargs: Any) -> AttributeDict:
        """
        Create inputs to launch immigrant from a code and a remote path on the associated computer.

        If POTCAR does not exist, the provided ``potential_family`` and
        ``potential_mapping`` are used to link potential to inputs. In this
        case, at least ``potential_family`` has to be provided. Unless
        ``potential_mapping``, this mapping is generated from structure, i.e.,

        ::

            potential_mapping = {element: element for element in structure.get_kind_names()}

        :param code: a Code instance for the code originally used.
        :param remote_workdir: Directory or folder name where VASP inputs and outputs are stored.
        :param settings: dict. This is used as the input port of VaspCalculation.
        :param potential_family: str. This will be obsolete at v3.0.
        :param potential_mapping: dict. This will be obsolete at v3.0.
        :param use_wavecar: bool. Try to read WAVECAR.
        :param use_chgcar bool. Try to read CHGCAR.

        """

        inputs = AttributeDict()
        inputs.code = code
        options = {'max_wallclock_seconds': 1, 'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 1}}
        inputs.metadata = AttributeDict()
        inputs.metadata.options = options
        inputs.remote_workdir = remote_workdir
        if 'settings' in kwargs:
            inputs.settings = orm.Dict(dict=kwargs['settings'])
        _remote_workdir = Path(remote_workdir)
        with cmp_get_transport(code.computer) as transport:
            with SandboxFolder() as sandbox:
                sandbox_path = Path(sandbox.abspath)
                transport.get(str(_remote_workdir / 'INCAR'), str(sandbox_path))
                transport.get(str(_remote_workdir / 'POSCAR'), str(sandbox_path))
                transport.get(str(_remote_workdir / 'POTCAR'), str(sandbox_path), ignore_nonexisting=True)
                transport.get(str(_remote_workdir / 'KPOINTS'), str(sandbox_path))
                inputs.parameters = get_incar_input(sandbox_path)
                inputs.structure = get_poscar_input(sandbox_path)
                inputs.kpoints = get_kpoints_input(sandbox_path, structure=inputs.structure)
                try:
                    inputs.potential = get_potcar_input(
                        sandbox_path,
                        structure=inputs.structure,
                        potential_family=kwargs.get('potential_family'),
                        potential_mapping=kwargs.get('potential_mapping'),
                    )
                except InputValidationError:
                    pass

                cls._add_inputs(transport, _remote_workdir, sandbox_path, inputs, **kwargs)

        return inputs

    @classmethod
    def get_builder_from_folder(cls, code: orm.Code, remote_workdir: str, **kwargs: Any) -> ProcessBuilder:
        """
        Create an immigrant builder from a code and a remote path on the associated computer.
        See more details in the docstring of ``get_inputs_from_folder``.
        """

        inputs = cls.get_inputs_from_folder(code, remote_workdir, **kwargs)
        builder = cls.get_builder()
        for key, val in inputs.items():
            builder[key] = val
        return builder

    @classmethod
    def _add_inputs(
        cls, transport: Any, remote_path: Path, sandbox_path: Path, inputs: AttributeDict, **kwargs: Any
    ) -> None:
        """Add some more inputs"""
        add_wavecar = kwargs.get('use_wavecar') or bool(inputs.parameters.get_dict().get('istart', 0))
        add_chgcar = kwargs.get('use_chgcar') or inputs.parameters.get_dict().get('icharg', -1) in [1, 11]
        if add_chgcar:
            transport.get(str(remote_path / 'CHGCAR'), str(sandbox_path))
            inputs.charge_density = get_chgcar_input(sandbox_path)
        if add_wavecar:
            transport.get(str(remote_path / 'WAVECAR'), str(sandbox_path))
            inputs.wavefunctions = get_wavecar_input(sandbox_path)


def get_incar_input(dir_path: Path) -> orm.Dict:
    """Create a node that contains the INCAR content."""
    from aiida_vasp.parsers.node_composer import NodeComposer  # noqa: PLC0415

    with open(str(dir_path / 'INCAR'), 'r', encoding='utf8') as handler:
        incar_parser = IncarParser(handler=handler)
    incar = incar_parser.get_quantity('incar')
    node = NodeComposer.compose_core_dict('core.dict', incar)

    return node


def get_poscar_input(dir_path: Path) -> orm.StructureData:
    """Create a node that contains the POSCAR content."""
    from aiida_vasp.parsers.node_composer import NodeComposer  # noqa: PLC0415

    with open(str(dir_path / 'POSCAR'), 'r', encoding='utf8') as handler:
        poscar_parser = PoscarParser(handler=handler)
    poscar = poscar_parser.get_quantity('poscar-structure')
    node = NodeComposer.compose_core_structure('core.structure', {'structure': poscar})

    return node


def get_potcar_input(
    dir_path: Path,
    structure: orm.StructureData | None = None,
    potential_family: str | None = None,
    potential_mapping: dict | None = None,
) -> dict[str, PotcarData]:
    """Read potentials from POTCAR or set it up from a structure."""
    local_potcar = dir_path / 'POTCAR'
    structure = structure or get_poscar_input(dir_path)
    potentials = {}
    if local_potcar.exists():
        potentials = MultiPotcarIo.read(str(local_potcar)).get_potentials_dict(structure)
        potentials = {kind: potentials[kind] for kind in potentials}
    elif potential_family:
        potentials = PotcarData.get_potcars_from_structure(structure, potential_family, mapping=potential_mapping)
    else:
        raise InputValidationError('no POTCAR found in remote folder and potential_family was not passed')

    return potentials


def get_kpoints_input(dir_path: Path, structure: orm.StructureData | None = None) -> orm.KpointsData:
    """Create a node that contains the KPOINTS content."""
    from aiida_vasp.parsers.node_composer import NodeComposer  # noqa: PLC0415

    structure = structure or get_poscar_input(dir_path)
    with open(str(dir_path / 'KPOINTS'), 'r', encoding='utf8') as handler:
        kpoints_parser = KpointsParser(handler=handler)
    kpoints = kpoints_parser.get_quantity('kpoints-kpoints')
    node = NodeComposer.compose_core_array_kpoints('core.array.kpoints', {'kpoints': kpoints})
    node.set_cell_from_structure(structure)

    return node


def get_chgcar_input(dir_path: Path) -> ChargedensityData:
    node = ChargedensityData(str(dir_path / 'CHGCAR'))

    return node


def get_wavecar_input(dir_path: Path) -> WavefunData:
    node = WavefunData(str(dir_path / 'WAVECAR'))

    return node



"""
Call script to calculate the total energies for one volume of standard silicon.

This particular call script set up a standard calculation that execute a calculation for
the fcc silicon structure.
"""
# pylint: disable=too-many-arguments
import numpy as np
from aiida import load_profile
from aiida.common.extendeddicts import AttributeDict
from aiida.engine import submit
from aiida.orm import Bool, Code, Str
from aiida.plugins import DataFactory, WorkflowFactory

load_profile()


def get_structure():
    """
    Set up Si primitive cell

    fcc Si:
       3.9
       0.5000000000000000    0.5000000000000000    0.0000000000000000
       0.0000000000000000    0.5000000000000000    0.5000000000000000
       0.5000000000000000    0.0000000000000000    0.5000000000000000
    Si
       1
    Cartesian
    0.0000000000000000  0.0000000000000000  0.0000000000000000

    """

    structure_data = DataFactory('structure')
    alat = 3.9
    lattice = np.array([[0.5, 0.5, 0], [0, 0.5, 0.5], [0.5, 0, 0.5]]) * alat
    structure = structure_data(cell=lattice)
    for pos_direct in [[0.0, 0.0, 0.0]]:
        pos_cartesian = np.dot(pos_direct, lattice)
        structure.append_atom(position=pos_cartesian, symbols='Si')
    return structure


def main(code_string, incar, kmesh, structure, potential_family, potential_mapping, options):
    """Main method to setup the calculation."""

    # First, we need to fetch the AiiDA datatypes which will
    # house the inputs to our calculation
    dict_data = DataFactory('dict')
    kpoints_data = DataFactory('array.kpoints')

    # Then, we set the workchain you would like to call
    workchain = WorkflowFactory('vasp.vasp')

    # And finally, we declare the options, settings and input containers
    settings = AttributeDict()
    inputs = AttributeDict()

    # Set inputs for the following WorkChain execution
    # Code
    inputs.code = Code.get_from_string(code_string)
    # Structure
    inputs.structure = structure
    # k-points grid density
    kpoints = kpoints_data()
    kpoints.set_kpoints_mesh(kmesh)
    inputs.kpoints = kpoints
    # Parameters
    inputs.parameters = dict_data(dict=incar)
    # Potential family and their mapping between element and potential type to use
    inputs.potential_family = Str(potential_family)
    inputs.potential_mapping = dict_data(dict=potential_mapping)
    # Options
    inputs.options = dict_data(dict=options)
    # Settings
    inputs.settings = dict_data(dict=settings)
    # Workchain related inputs, in this case, give more explicit output to report
    inputs.verbose = Bool(True)
    # Submit the workchain with the set inputs
    submit(workchain, **inputs)


if __name__ == '__main__':
    # Code_string is chosen among the list given by 'verdi code list'
    CODE_STRING = 'vasp@mycluster'

    # INCAR equivalent
    # Set input parameters
    INCAR = {'incar': {'istart': 0, 'icharg': 2, 'encut': 240, 'ismear': 0, 'sigma': 0.1}}

    # KPOINTS equivalent
    # Set kpoint mesh
    KMESH = [11, 11, 11]

    # POTCAR equivalent
    # Potential_family is chosen among the list given by
    # 'verdi data vasp.potcar listfamilies'
    POTENTIAL_FAMILY = 'PBE.54'
    # The potential mapping selects which potential to use, here we use the standard
    # for silicon, this could for instance be {'Si': 'Si_GW'} to use the GW ready
    # potential instead
    POTENTIAL_MAPPING = {'Si': 'Si'}

    # Jobfile equivalent
    # In options, we typically set scheduler options.
    # See https://aiida.readthedocs.io/projects/aiida-core/en/latest/scheduler/index.html
    # AttributeDict is just a special dictionary with the extra benefit that
    # you can set and get the key contents with mydict.mykey, instead of mydict['mykey']
    OPTIONS = AttributeDict()
    OPTIONS.account = ''
    OPTIONS.qos = ''
    OPTIONS.resources = {'num_machines': 1, 'num_mpiprocs_per_machine': 1}
    OPTIONS.queue_name = ''
    OPTIONS.max_wallclock_seconds = 3600
    OPTIONS.max_memory_kb = 2000000

    # POSCAR equivalent
    # Set the silicon structure
    STRUCTURE = get_structure()

    main(CODE_STRING, INCAR, KMESH, STRUCTURE, POTENTIAL_FAMILY, POTENTIAL_MAPPING, OPTIONS)



"""Test aiida_parameters."""

# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import,no-member, import-outside-toplevel
import re

import pytest
from aiida.common.extendeddicts import AttributeDict
from aiida.plugins import DataFactory

from aiida_vasp.assistant.parameters import _DEFAULT_OVERRIDE_NAMESPACE, ParametersMassage, inherit_and_merge_parameters


@pytest.fixture
def init_relax_parameters():
    """Fixture for a set of general input parameters for relaxation."""
    general_parameters = AttributeDict()
    general_parameters.relax = AttributeDict()
    general_parameters.relax.algo = 'cg'
    general_parameters.relax.force_cutoff = 0.01
    general_parameters.relax.steps = 60
    general_parameters.relax.positions = True
    general_parameters.relax.shape = True
    general_parameters.relax.volume = True

    return general_parameters


@pytest.fixture
def init_smearing_parameters():
    """Fixture for a set of general input parameters regarding Fermi integral smearing."""
    general_parameters = AttributeDict()
    general_parameters.smearing = AttributeDict()
    general_parameters.smearing.gaussian = True

    return general_parameters


@pytest.fixture
def init_charge_parameters():
    """Fixture for a set of general input parameters regarding how to construct initial charge density."""
    general_parameters = AttributeDict()
    general_parameters.charge = AttributeDict()
    general_parameters.charge.constant_charge = True

    return general_parameters


@pytest.fixture
def init_bands_parameters():
    """Fixture for a set of general input parameters for band structure calculations."""
    general_parameters = AttributeDict()
    general_parameters.bands = AttributeDict()
    general_parameters.bands.decompose_bands = False
    general_parameters.bands.decompose_wave = False

    return general_parameters


def test_override_namespace():
    """Test that the default override namespace is incar."""
    assert _DEFAULT_OVERRIDE_NAMESPACE == 'incar'


def test_relax_parameters_all_set(init_relax_parameters):
    """Test all standard relaxation parameters are set."""
    massager = ParametersMassage(init_relax_parameters)
    parameters = massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE]
    assert parameters.ediffg == pytest.approx(-0.01)
    assert parameters.ibrion == 2
    assert parameters.nsw == 60
    assert parameters.isif == 3


def test_additional_override_namespaces(init_relax_parameters):  # pylint: disable=invalid-name
    """Test that we can supply additional override namespaces and that they are unmodified in the massager."""
    init_relax_parameters.myspace = AttributeDict({'myspaceparameter': 1})
    init_relax_parameters.yourspace = AttributeDict({'yourspaceparameter': 1})
    massager = ParametersMassage(
        init_relax_parameters, settings=AttributeDict({'additional_override_namespaces': ['myspace', 'yourspace']})
    )
    assert massager.parameters.myspace.myspaceparameter == 1
    assert massager.parameters.yourspace.yourspaceparameter == 1


def test_catch_invalid_tags(init_relax_parameters):
    """Test to see if the massager accepts an invalid tag."""
    init_relax_parameters[_DEFAULT_OVERRIDE_NAMESPACE] = AttributeDict()
    init_relax_parameters[_DEFAULT_OVERRIDE_NAMESPACE].smear = 1  # This is an invalid tag
    matching_string = re.compile(r'^The supplied key: smear is not a support VASP parameter.$')
    with pytest.raises(ValueError, match=matching_string):
        _ = ParametersMassage(init_relax_parameters)


def test_relax_multiple_cutoffs(init_relax_parameters):
    """Test if the massager raise exception if both energy and force cutoff is supplied."""
    init_relax_parameters.relax.energy_cutoff = 0.01
    matching_string = re.compile(
        r'^User supplied both a force and an energy cutoff for the relaxation. Please select.$'
    )
    with pytest.raises(ValueError, match=matching_string):
        _ = ParametersMassage(init_relax_parameters)


def test_allowed_namespaces(init_relax_parameters):
    """Test if the massager raise exception if an unsupported namespace is supplied."""
    init_relax_parameters.not_allowed_namespace = 0.1
    matching_string = re.compile(r'^The supplied namespace: not_allowed_namespace is not supported.$')
    with pytest.raises(ValueError, match=matching_string):
        _ = ParametersMassage(init_relax_parameters)


def test_relax_parameters_cutoff(init_relax_parameters):
    """Test no provided cutoff yields no set ediffg."""
    del init_relax_parameters.relax.force_cutoff
    massager = ParametersMassage(init_relax_parameters)
    parameters = massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE]
    with pytest.raises(AttributeError):
        _ = parameters.ediffg


def test_relax_parameters_no_algo(init_relax_parameters):
    """Test no provided algo tag."""
    del init_relax_parameters.relax.algo
    massager = ParametersMassage(init_relax_parameters)
    with pytest.raises(AttributeError):
        _ = massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].ibrion


def test_relax_parameters_vol_shape(init_relax_parameters):
    """Test volume and shape relaxation combinations."""
    del init_relax_parameters.relax.positions
    massager = ParametersMassage(init_relax_parameters)
    parameters = massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE]
    assert parameters.isif == 6


def test_relax_parameters_pos_shape(init_relax_parameters):
    """Test position and shape relxation combinations."""
    del init_relax_parameters.relax.volume
    massager = ParametersMassage(init_relax_parameters)
    parameters = massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE]
    assert parameters.isif == 4


def test_relax_parameters_vol(init_relax_parameters):
    """Test only volume relaxation."""
    del init_relax_parameters.relax.positions
    del init_relax_parameters.relax.shape
    massager = ParametersMassage(init_relax_parameters)
    parameters = massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE]
    assert parameters.isif == 7


def test_relax_parameters_pos(init_relax_parameters):
    """Test only position relaxation."""
    del init_relax_parameters.relax.volume
    del init_relax_parameters.relax.shape
    massager = ParametersMassage(init_relax_parameters)
    parameters = massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE]
    assert parameters.isif == 2


def test_relax_parameters_shape(init_relax_parameters):
    """Test only shape relaxation."""
    del init_relax_parameters.relax.volume
    del init_relax_parameters.relax.positions
    massager = ParametersMassage(init_relax_parameters)
    parameters = massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE]
    assert parameters.isif == 5


def test_relax_parameters_nothing(init_relax_parameters):
    """Test if no relaxation parameters for volume, positions and shape are given."""
    del init_relax_parameters.relax.volume
    del init_relax_parameters.relax.positions
    del init_relax_parameters.relax.shape
    massager = ParametersMassage(init_relax_parameters)
    parameters = massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE]
    assert parameters == AttributeDict()


def test_parameters_override(init_relax_parameters):
    """Test what happens if we override a parameters."""
    value = 1  # should from init_relax_parameters be 3, verified in test_relax_parameters_all_set
    init_relax_parameters[_DEFAULT_OVERRIDE_NAMESPACE] = AttributeDict()
    # Add this override to the incar namespace which is considered the override namespace before reaching
    # ParameterMassager
    init_relax_parameters[_DEFAULT_OVERRIDE_NAMESPACE].isif = value
    massager = ParametersMassage(init_relax_parameters)
    parameters = massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE]
    assert parameters.isif == value


def test_smearing_parameters():
    """Test smearing parameters."""
    parameters = AttributeDict()
    parameters.smearing = AttributeDict()
    parameters.smearing.gaussian = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].ismear == 0
    parameters.smearing.gaussian = False
    parameters.smearing.fermi = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].ismear == -1
    parameters.smearing.fermi = False
    parameters.smearing.tetra = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].ismear == -5
    parameters.smearing.tetra = False
    parameters.smearing.mp = 4
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].ismear == 4


def test_charge_parameters():
    """Test charge parameters."""
    parameters = AttributeDict()
    parameters.charge = AttributeDict()
    parameters.charge.from_wave = True
    massager = ParametersMassage(parameters)
    parameters.charge.from_wave = False
    parameters.charge.from_charge = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].icharg == 1
    parameters.charge.from_charge = False
    parameters.charge.from_atomic = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].icharg == 2
    parameters.charge.from_atomic = False
    parameters.charge.from_potential = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].icharg == 4
    parameters.charge.from_potential = False
    parameters.charge.constant_charge = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].icharg == 11
    parameters.charge.constant_charge = False
    parameters.charge.constant_atomic = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].icharg == 12


def test_orbital_projections():  # pylint: disable=too-many-statements
    """Test the parameters associated with orbital projections."""
    parameters = AttributeDict()
    parameters.bands = AttributeDict()
    parameters.bands.decompose_wave = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].lorbit == 5
    parameters.bands.decompose_wave = False
    parameters.bands.decompose_bands = True
    parameters.bands.decompose_auto = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].lorbit == 14
    parameters.bands.decompose_auto = False
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].lorbit == 10
    parameters.bands.lm = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].lorbit == 11
    parameters.bands.phase = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].lorbit == 12
    parameters.bands.lm = False
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].lorbit == 12

    # Now do the once with a Wigner-Seitz radius supplied
    parameters.bands.wigner_seitz_radius = [2.0]
    parameters.bands.lm = False
    parameters.bands.phase = False
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].lorbit == 0
    assert int(massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].rwigs[0]) == 2
    parameters.bands.lm = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].lorbit == 1
    parameters.bands.phase = True
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].lorbit == 2
    parameters.bands.lm = False
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].lorbit == 2

    # Should raise ValueError if Wigner-Seitz radius is not defined as a list.
    parameters.bands.wigner_seitz_radius = 2.0
    with pytest.raises(ValueError):
        massager = ParametersMassage(parameters)


def test_vasp_parameter_override(init_relax_parameters):
    """Test of the override functionality works as intended."""
    init_relax_parameters[_DEFAULT_OVERRIDE_NAMESPACE] = AttributeDict()
    # Redefine to from 3 to 0.
    init_relax_parameters[_DEFAULT_OVERRIDE_NAMESPACE].isif = 0
    massager = ParametersMassage(init_relax_parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].isif == 0


def test_inherit_and_merge():
    """Test the inherit and merge functionality for the parameters and inputs."""

    inputs = AttributeDict()
    inputs.bands = AttributeDict()
    inputs.bands.somekey = DataFactory('core.bool')(True)
    inputs.relax = AttributeDict()
    inputs.relax.somekey = DataFactory('core.bool')(True)
    inputs.smearing = AttributeDict()
    inputs.smearing.somekey = DataFactory('core.bool')(True)
    inputs.charge = AttributeDict()
    inputs.charge.somekey = DataFactory('core.bool')(True)
    inputs.converge = AttributeDict()
    inputs.converge.somekey = DataFactory('core.bool')(True)
    inputs.electronic = AttributeDict()
    inputs.electronic.somekey = DataFactory('core.bool')(True)
    inputs.dynamics = AttributeDict()
    inputs.dynamics.somekey = DataFactory('core.bool')(True)
    # Check that parameters does not have to be present
    parameters = inherit_and_merge_parameters(inputs)
    # Check that an empty parameters is allowed
    inputs.parameters = DataFactory('core.dict')(dict={})
    parameters = inherit_and_merge_parameters(inputs)
    test_parameters = AttributeDict(
        {
            'electronic': AttributeDict({'somekey': True}),
            'bands': AttributeDict({'somekey': True}),
            'smearing': AttributeDict({'somekey': True}),
            'charge': AttributeDict({'somekey': True}),
            'relax': AttributeDict({'somekey': True}),
            'converge': AttributeDict({'somekey': True}),
            'dynamics': AttributeDict({'somekey': True}),
        }
    )
    assert parameters == test_parameters
    # Test ignored
    inputs.ignored = AttributeDict()
    inputs.ignored.ignored = DataFactory('core.bool')(True)
    parameters = inherit_and_merge_parameters(inputs)
    assert parameters == test_parameters
    # Test to override inputs.bands.somekey
    inputs.parameters = DataFactory('core.dict')(dict={'bands': {'somekey': False}})
    parameters = inherit_and_merge_parameters(inputs)
    test_parameters.bands.somekey = False
    assert parameters == test_parameters


def test_unsupported_fail_override():
    """Test that any supplied unsupported parameters in the regular parameters dictionary yield error."""
    parameters = AttributeDict()
    parameters[_DEFAULT_OVERRIDE_NAMESPACE] = AttributeDict()
    parameters[_DEFAULT_OVERRIDE_NAMESPACE].not_valid = 200
    matching_string = re.compile(r'^The supplied key: not_valid is not a support VASP parameter.$')
    with pytest.raises(ValueError, match=matching_string):
        _ = ParametersMassage(parameters)


def test_unsupported_parameters_in_unsupported_namespace():  # pylint: disable=invalid-name
    """Test that it is possibly to supply unsupported parameters in the incar namespace if the configuration is also
    supplied."""
    parameters = AttributeDict()
    parameters[_DEFAULT_OVERRIDE_NAMESPACE] = AttributeDict()
    parameters[_DEFAULT_OVERRIDE_NAMESPACE].not_valid = 200
    massager = ParametersMassage(
        parameters,
        unsupported_parameters={
            'not_valid': {'default': 1.0, 'description': 'Something', 'type': float, 'values': [1.0, 2.0]}
        },
    )
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].not_valid == 200


def test_skip_parameters_validate():  # pylint: disable=invalid-name
    """Test that it is possibly to completely by-pass parameters checking."""
    parameters = AttributeDict()
    parameters[_DEFAULT_OVERRIDE_NAMESPACE] = AttributeDict()
    parameters[_DEFAULT_OVERRIDE_NAMESPACE].not_valid = 200
    massager = ParametersMassage(parameters, skip_parameters_validation=True)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].not_valid == 200


def test_pwcutoff_to_encut():
    """Test that the pwcutoff is converted to encut."""
    parameters = AttributeDict()
    parameters.electronic = AttributeDict()
    parameters.electronic.pwcutoff = 200
    massager = ParametersMassage(parameters)
    assert massager.parameters[_DEFAULT_OVERRIDE_NAMESPACE].encut == parameters.electronic.pwcutoff



"""
Unit tests for vasp.potcar command family.
"""

import os
import subprocess
from pathlib import Path

import pytest
from aiida.common import AttributeDict
from aiida.orm import Group, Node
from aiida.orm.querybuilder import QueryBuilder
from click.testing import CliRunner

from aiida_vasp.commands.potcar import potcar
from aiida_vasp.data.potcar import PotcarData, PotcarGroup


@pytest.fixture
def cmd_params(temp_pot_folder, potcar_family_name):
    """Common building blocks for ``uploadfamily`` calls."""
    params = AttributeDict()
    params.POTCAR_PATH = str(temp_pot_folder)
    params.FAMILY_NAME = potcar_family_name
    params.PATH_OPTION = f'--path={params.POTCAR_PATH}'
    params.NAME_OPTION = f'--name={params.FAMILY_NAME}'
    params.DESC_OPTION = '--description="This is a test POTCAR family"'
    return params


def run_cmd(command=None, args=None, **kwargs):
    """Run verdi data vasp.potcar <command> [args]."""
    runner = CliRunner()
    params = args or []
    if command:
        params.insert(0, command)
    return runner.invoke(potcar, params, **kwargs)


def test_no_subcmd():
    result = run_cmd('--help')
    assert result.exception is None


def test_uploadfamily_withpath(aiida_profile_clean, cmd_params):
    """Upload the test potcar family and check it is there."""

    result = run_cmd(
        'uploadfamily',
        [cmd_params.PATH_OPTION, cmd_params.NAME_OPTION, cmd_params.DESC_OPTION],
    )

    potcar_cls = PotcarData

    assert not result.exception
    assert potcar_cls.exists(element='In')
    assert potcar_cls.exists(element='Ga')
    assert [g.label for g in potcar_cls.get_potcar_groups()] == [cmd_params.FAMILY_NAME]


def test_uploadfamily_tar(aiida_profile_clean, cmd_params):
    """Give a tar file as the source."""
    path_option = f'--path={Path(cmd_params.POTCAR_PATH) / "Ga.tar"!s}'
    result = run_cmd('uploadfamily', [path_option, cmd_params.NAME_OPTION, cmd_params.DESC_OPTION])
    potcar_cls = PotcarData

    assert not result.exception
    assert potcar_cls.exists(element='Ga')
    assert [g.label for g in potcar_cls.get_potcar_groups()] == [cmd_params.FAMILY_NAME]


def test_uploadfamily_inworkdir(aiida_profile_clean, cmd_params):
    """Upload the test potcar family from the working env."""

    potcar_dir = Path(cmd_params.POTCAR_PATH)
    old_work_dir = Path().cwd()
    os.chdir(str(potcar_dir))
    assert str(potcar_dir) == str(Path().cwd())

    result = run_cmd('uploadfamily', [cmd_params.NAME_OPTION, cmd_params.DESC_OPTION])

    potcar_cls = PotcarData

    assert not result.exception
    assert potcar_cls.exists(element='In')
    assert [g.label for g in potcar_cls.get_potcar_groups()] == [cmd_params.FAMILY_NAME]

    os.chdir(str(old_work_dir))

    assert str(old_work_dir) == str(Path().cwd())


def test_uploadfamily_again(aiida_profile_clean, upload_potcar, cmd_params):
    """
    Re-upload a potcar family.

    Test:
        * Does not require description
        * Must succeed
        * Adds no nodes
        * Adds no groups
    """

    node_qb = QueryBuilder(path=[Node])
    node_count = node_qb.count()
    group_qb = QueryBuilder(path=[Group])
    group_count = group_qb.count()

    result = run_cmd('uploadfamily', [cmd_params.PATH_OPTION, cmd_params.NAME_OPTION])

    assert not result.exception

    node_qb = QueryBuilder(path=[Node])
    assert node_count == node_qb.count()
    group_qb = QueryBuilder(path=[Group])
    assert group_count == group_qb.count()


def test_uploadfamily_dryrun(aiida_profile_clean, cmd_params):
    """Make sure --dry-run does not affect the db."""
    node_qb = QueryBuilder(path=[Node])
    node_count = node_qb.count()
    group_qb = QueryBuilder(path=[Group])
    group_count = group_qb.count()

    result = run_cmd(
        'uploadfamily',
        [
            cmd_params.PATH_OPTION,
            cmd_params.NAME_OPTION,
            cmd_params.DESC_OPTION,
            '--dry-run',
        ],
    )

    assert not result.exception

    node_qb = QueryBuilder(path=[Node])
    assert node_count == node_qb.count()
    group_qb = QueryBuilder(path=[Group])
    assert group_count == group_qb.count()


def test_listfamilies_existence():
    """Make sure the listfamilies subcommand exists."""
    result = run_cmd('listfamilies')
    assert not result.exception
    assert result.output


def test_listfamilies_nofilter(aiida_profile_clean, upload_potcar, potcar_family_name):
    """Test typical usecases without filtering."""
    result = run_cmd('listfamilies')
    assert not result.exception
    assert potcar_family_name in result.output

    family_group = PotcarData.get_potcar_group(potcar_family_name)
    result = run_cmd('listfamilies', ['--description'])
    assert not result.exception
    assert 'Description' in result.output
    assert family_group.description in result.output


def test_listfamilies_filtering(aiida_profile_clean, upload_potcar, potcar_family_name):
    """Test filtering families by elements & symbols."""
    result = run_cmd('listfamilies', ['--element', 'In', '--element', 'As'])
    assert potcar_family_name in result.output

    result = run_cmd('listfamilies', ['--element', 'In', '--element', 'U235'])
    assert potcar_family_name not in result.output

    result = run_cmd('listfamilies', ['--symbol', 'In_d'])
    assert potcar_family_name in result.output

    result = run_cmd('listfamilies', ['--symbol', 'In_d', '--symbol', 'In_s'])
    assert potcar_family_name not in result.output

    result = run_cmd('listfamilies', ['--symbol', 'In_d', '--element', 'As'])
    assert potcar_family_name in result.output

    result = run_cmd('listfamilies', ['--symbol', 'In_d', '--element', 'U235'])
    assert potcar_family_name not in result.output


def test_exportfamilies(aiida_profile_clean, upload_potcar, potcar_family_name, tmp_path):
    """Test exporting potcar family."""
    result = run_cmd('exportfamily', ['--name', potcar_family_name, '--path', str(tmp_path)])
    assert not result.exception
    export_path = tmp_path / potcar_family_name
    assert export_path.is_dir()
    assert export_path.exists()

    new_dir = tmp_path / 'new_dir'
    result = run_cmd('exportfamily', ['--dry-run', '--name', potcar_family_name, '--path', new_dir])
    assert not result.exception
    assert not new_dir.exists()

    result = run_cmd('exportfamily', ['--as-archive', '--name', potcar_family_name, '--path', tmp_path])
    assert not result.exception
    name = potcar_family_name + '.tar.gz'
    export_path = tmp_path / name
    assert export_path.is_file()
    assert export_path.exists()

    new_arch = tmp_path / 'export.tar.gz'
    result = run_cmd(
        'exportfamily',
        ['--dry-run', '--as-archive', '--name', potcar_family_name, '--path', new_arch],
    )
    assert not result.exception
    assert not new_arch.exists()


def test_call_from_vasp():
    """Test if the verdi potcar data command works."""

    output = subprocess.check_output(['verdi', 'data', 'vasp.potcar', '--help'], universal_newlines=True)
    assert 'Usage: verdi data vasp.potcar' in output  # pylint: disable=unsupported-membership-test


def test_migrate_command(aiida_profile_clean, legacy_potcar_family):
    """Test the migration command"""

    legacy_name, legacy_group_class = legacy_potcar_family
    legacy_group = legacy_group_class.collection.get(label=legacy_name)
    run_cmd('migratefamilies')
    migrated = PotcarGroup.collection.get(label=legacy_name)

    uuids_original = {node.uuid for node in legacy_group.nodes}
    uuids_migrated = {node.uuid for node in migrated.nodes}
    assert uuids_migrated == uuids_original



import numpy as np
from aiida.engine import submit
from aiida.manage.configuration import load_profile
from aiida.orm import Bool, Code, Str
from aiida.plugins import DataFactory, WorkflowFactory

load_profile()


def launch_aiida(
    structure,
    code_string,
    options,
    potential_family,
    potential_mapping,
    label='SiC VASP calculation',
):
    Dict = DataFactory('dict')
    KpointsData = DataFactory('array.kpoints')

    incar_dict = {
        'PREC': 'Accurate',
        'IBRION': -1,
        'EDIFF': 1e-8,
        'NELMIN': 5,
        'NELM': 100,
        'ENCUT': 500,
        'IALGO': 38,
        'ISMEAR': 0,
        'SIGMA': 0.01,
        'LREAL': False,
        'LCHARG': False,
        'LWAVE': False,
    }

    kpoints = KpointsData()
    kpoints.set_kpoints_mesh([6, 6, 4], offset=[0, 0, 0.5])

    parser_settings = {'add_energies': True, 'add_forces': True, 'add_stress': True}

    code = Code.get_from_string(code_string)
    Workflow = WorkflowFactory('vasp.vasp')
    builder = Workflow.get_builder()
    builder.code = code
    builder.parameters = Dict(dict={'incar': incar_dict})
    builder.structure = structure
    builder.settings = Dict(dict={'parser_settings': parser_settings})
    builder.potential_family = Str(potential_family)
    builder.potential_mapping = Dict(dict=potential_mapping)
    builder.kpoints = kpoints
    builder.options = Dict(dict=options)
    builder.metadata.label = label
    builder.metadata.description = label
    builder.clean_workdir = Bool(False)
    node = submit(builder)
    return node


def get_structure_SiC():
    """Set up SiC cell

    Si C
       1.0
         3.0920072935808083    0.0000000000000000    0.0000000000000000
        -1.5460036467904041    2.6777568649277486    0.0000000000000000
         0.0000000000000000    0.0000000000000000    5.0733470000000001
     Si C
       2   2
    Direct
       0.3333333333333333  0.6666666666666665  0.4995889999999998
       0.6666666666666667  0.3333333333333333  0.9995889999999998
       0.3333333333333333  0.6666666666666665  0.8754109999999998
       0.6666666666666667  0.3333333333333333  0.3754109999999997

    """

    StructureData = DataFactory('structure')
    a = 3.092
    c = 5.073
    lattice = [[a, 0, 0], [-a / 2, a / 2 * np.sqrt(3), 0], [0, 0, c]]
    structure = StructureData(cell=lattice)
    for pos_direct, symbol in zip(
        (
            [1.0 / 3, 2.0 / 3, 0],
            [2.0 / 3, 1.0 / 3, 0.5],
            [1.0 / 3, 2.0 / 3, 0.375822],
            [2.0 / 3, 1.0 / 3, 0.875822],
        ),
        ('Si', 'Si', 'C', 'C'),
    ):
        pos_cartesian = np.dot(pos_direct, lattice)
        structure.append_atom(position=pos_cartesian, symbols=symbol)
    return structure


def main(code_string, options, potential_family, potential_mapping):
    structure = get_structure_SiC()
    launch_aiida(structure, code_string, options, potential_family, potential_mapping)


if __name__ == '__main__':
    code_string = 'vasp@mycluster'
    options = {
        'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 8},
        'account': '',
        'max_memory_kb': 2000000,
        'max_wallclock_seconds': 3600,
    }
    potential_family = 'PBE.54'
    potential_mapping = {'Si': 'Si', 'C': 'C'}
    main(code_string, options, potential_family, potential_mapping)



"""
Fixtures related to data.

-------------------------
Here different pytest fixtures are set up. They typically contain computers,
VASP inputs etc. which you would need to mock a VASP job in this plugin. It
also contains the set up of the mock VASP executable, which is used to test the
workchains.
"""

import pytest
from aiida import orm
from aiida.plugins import DataFactory

from aiida_vasp.data.potcar import PotcarData, PotcarFileData


@pytest.fixture
def vasp2w90_params(aiida_profile, vasp_params):
    vasp_params_data = vasp_params()
    incar_data = orm.Dict(dict=vasp_params_data.code.get_dict().update({'lwannier90': True}))
    return incar_data


@pytest.fixture
def potcar_node_pair(aiida_profile, data_path):
    """Create a POTCAR node pair."""
    potcar_path = data_path('potcar', 'As', 'POTCAR')
    potcar_file_node = PotcarFileData(file=potcar_path)
    potcar_file_node.store()
    return {'file': potcar_file_node, 'potcar': PotcarData.find_one(symbol='As')}


@pytest.fixture()
def ref_retrieved(data_path):
    """Fixture: retrieved directory from an NSCF vasp run."""
    retrieved = DataFactory('core.folder')()
    retrieved.put_object_from_tree(path=data_path('basic_run'))
    return retrieved
